"""
import liberator

from netharn.data.batch_samplers import PatchedBatchSampler
from netharn.data.data_containers import ItemContainer
from netharn.data.data_containers import container_collate
from netharn.data.batch_samplers import PatchedRandomSampler
from netharn.data.batch_samplers import SubsetSampler
lib = liberator.Liberator()
lib.add_dynamic(SubsetSampler)
lib.add_dynamic(PatchedRandomSampler)
lib.add_dynamic(container_collate)
lib.add_dynamic(ItemContainer)
lib.add_dynamic(PatchedBatchSampler)
lib.add_dynamic(padded_colate)
lib.expand(['netharn'])
print(lib.current_sourcecode())
"""

import torch
import ubelt as ub
import torch.nn.functional as F
import numpy as np
import re
import torch.utils.data as torch_data

try:
    import collections.abc as container_abcs
    from six import string_types as string_classes
    from six import integer_types as int_classes
except Exception:
    from torch._six import container_abcs
    from torch._six import string_classes, int_classes


class PatchedBatchSampler(torch.utils.data.sampler.BatchSampler, ub.NiceRepr):
    """
    A modification of the standard torch BatchSampler that allows for
    specification of ``num_batches=auto``

    Example:
        >>> data_source = torch.arange(64)
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=None)
        >>> batch_size = 10
        >>> drop_last = False
        >>> num_batches = 'auto'
        >>> batch_sampler = PatchedBatchSampler(sampler, batch_size, drop_last, num_batches)
        >>> assert len(list(batch_sampler)) == 7 == len(batch_sampler)
        >>> batch_sampler = PatchedBatchSampler(sampler, batch_size, drop_last, 3)
        >>> assert len(list(batch_sampler)) == 3 == len(batch_sampler)
        >>> batch_sampler = PatchedBatchSampler(sampler, batch_size, drop_last, 1)
        >>> assert len(list(batch_sampler)) == 1 == len(batch_sampler)
    """
    def __init__(self, sampler, batch_size, drop_last, num_batches='auto'):
        super().__init__(sampler, batch_size, drop_last)
        self.num_batches = num_batches

    def __len__(self):
        if self.drop_last:
            max_num_batches = len(self.sampler) // self.batch_size  # type: ignore
        else:
            max_num_batches = (len(self.sampler) + self.batch_size - 1) // self.batch_size  # type: ignore
        if self.num_batches == 'auto':
            num_batches = max_num_batches
        else:
            num_batches = min(max_num_batches, self.num_batches)
        return num_batches

    def __iter__(self):
        num_batches = len(self)
        for bx, batch in zip(range(num_batches), super().__iter__()):
            yield batch


class ItemContainer(ub.NiceRepr):
    """
    A container for uncollated items that defines a specific collation
    strategy. Based on mmdetections ItemContainer.
    """

    def __init__(
        self,
        data,
        stack=False,
        padding_value=-1,
        cpu_only=False,
        pad_dims=2
    ):
        self._data = data
        assert pad_dims in [None, 1, 2, 3]
        self.meta = {
            'stack': stack,
            'padding_value': padding_value,
            'cpu_only': cpu_only,
            'pad_dims': pad_dims,
        }

    @property
    def nestshape(self):
        return nestshape(self.data)

    def __nice__(self):
        try:
            shape_repr = ub.repr2(self.nestshape, nl=-2)
            return 'nestshape(data)={}'.format(shape_repr)
        except Exception:
            return object.__repr__(self)
            # return super().__repr__()

    @classmethod
    def demo(cls, key='img', rng=None, **kwargs):
        """
        Create data for tests

        Example:
            >>> print(ItemContainer.demo('img'))
            >>> print(ItemContainer.demo('labels'))
            >>> print(ItemContainer.demo('box'))

        """
        import kwarray
        rng = kwarray.ensure_rng(rng)
        if key == 'img':
            shape = kwargs.get('shape', (3, 512, 512))
            data = rng.rand(*shape).astype(np.float32)
            data = torch.from_numpy(data)
            self = cls(data, stack=True)
        elif key == 'labels':
            n = rng.randint(0, 10)
            data = rng.randint(0, 10, n)
            data = torch.from_numpy(data)
            self = cls(data, stack=False)
        elif key == 'box':
            n = rng.randint(0, 10)
            data = rng.rand(n, 4)
            data = torch.from_numpy(data)
            self = cls(data, stack=False)
        else:
            raise KeyError(key)
        return self

    def __getitem__(self, index):
        assert self.stack, 'can only index into stackable items'
        cls = self.__class__
        return cls(self.data[index], **self.meta)

    @property
    def data(self):
        return self._data

    @property
    def datatype(self):
        if isinstance(self.data, torch.Tensor):
            return self.data.type()
        else:
            return type(self.data)

    @property
    def cpu_only(self):
        return self.meta['cpu_only']

    @property
    def stack(self):
        return self.meta['stack']

    @property
    def padding_value(self):
        return self.meta['padding_value']

    @property
    def pad_dims(self):
        return self.meta['pad_dims']

    def size(self, *args, **kwargs):
        return self.data.size(*args, **kwargs)

    @property
    def shape(self):
        return self.data.shape

    def dim(self):
        return self.data.dim()

    @classmethod
    def _collate(cls, inbatch, num_devices=None):
        """
        Collates a sequence of DataContainers

        Args:
            inbatch (Sequence[ItemContainer]): datacontainers with the same
                parameters.

            num_devices (int): number of groups, if None, then uses one group.

        Example:
            >>> print('Collate Image ItemContainer')
            >>> inbatch = [ItemContainer.demo('img') for _ in range(5)]
            >>> print('inbatch = {}'.format(ub.repr2(inbatch)))
            >>> result = ItemContainer._collate(inbatch, num_devices=2)
            >>> print('result1 = {}'.format(ub.repr2(result, nl=1)))
            >>> result = ItemContainer._collate(inbatch, num_devices=1)
            >>> print('result2 = {}'.format(ub.repr2(result, nl=1)))
            >>> result = ItemContainer._collate(inbatch, num_devices=None)
            >>> print('resultN = {}'.format(ub.repr2(result, nl=1)))

            >>> print('Collate Label ItemContainer')
            >>> inbatch = [ItemContainer.demo('labels') for _ in range(5)]
            >>> print('inbatch = {}'.format(ub.repr2(inbatch, nl=1)))
            >>> result = ItemContainer._collate(inbatch, 1)
            >>> print('result1 = {}'.format(ub.repr2(result, nl=1)))
            >>> result = ItemContainer._collate(inbatch, 2)
            >>> print('result2 = {}'.format(ub.repr2(result, nl=1)))
            >>> result = ItemContainer._collate(inbatch, None)
            >>> print('resultN = {}'.format(ub.repr2(result, nl=1)))
        """
        item0 = inbatch[0]
        bsize = len(inbatch)
        if num_devices is None:
            num_devices = 1

        samples_per_device = int(np.ceil(bsize / num_devices))

        # assert bsize % samples_per_device == 0
        stacked = []
        if item0.cpu_only:
            # chunking logic
            stacked = []
            for i in range(0, bsize, samples_per_device):
                stacked.append(
                    [sample.data for sample in inbatch[i:i + samples_per_device]])

        elif item0.stack:
            for i in range(0, bsize, samples_per_device):
                item = inbatch[i]
                pad_dims_ = item.pad_dims
                assert isinstance(item.data, torch.Tensor)

                if pad_dims_ is not None:
                    # Note: can probably reimplement this using padded collate
                    # logic
                    ndim = item.dim()
                    assert ndim > pad_dims_
                    max_shape = [0 for _ in range(pad_dims_)]
                    for dim in range(1, pad_dims_ + 1):
                        max_shape[dim - 1] = item.shape[-dim]
                    for sample in inbatch[i:i + samples_per_device]:
                        for dim in range(0, ndim - pad_dims_):
                            assert item.shape[dim] == sample.shape[dim]
                        for dim in range(1, pad_dims_ + 1):
                            max_shape[dim - 1] = max(max_shape[dim - 1], sample.shape[-dim])
                    padded_samples = []
                    for sample in inbatch[i:i + samples_per_device]:
                        pad = [0 for _ in range(pad_dims_ * 2)]
                        for dim in range(1, pad_dims_ + 1):
                            pad[2 * dim - 1] = max_shape[dim - 1] - sample.shape[-dim]
                        padded_samples.append(
                            F.pad(sample.data, pad, value=sample.padding_value))
                    stacked.append(default_collate(padded_samples))

                elif pad_dims_ is None:
                    stacked.append(
                        default_collate([
                            sample.data
                            for sample in inbatch[i:i + samples_per_device]
                        ]))
                else:
                    raise ValueError(
                        'pad_dims should be either None or integers (1-3)')

        else:
            for i in range(0, bsize, samples_per_device):
                stacked.append(
                    [sample.data for sample in inbatch[i:i + samples_per_device]])
        result = BatchContainer(stacked, **item0.meta)
        return result


def nestshape(data):
    """
    Examine nested shape of the data

    Example:
        >>> data = [np.arange(10), np.arange(13)]
        >>> nestshape(data)
        [(10,), (13,)]

    Ignore:
        >>> # xdoctest: +REQUIRES(module:mmdet)
        >>> from mmdet.core.mask.structures import *  # NOQA
        >>> masks = [
        >>>     [ np.array([0, 0, 10, 0, 10, 10., 0, 10, 0, 0]) ],
        >>>     [ np.array([0, 0, 10, 0, 10, 10., 0, 10, 5., 5., 0, 0]) ]
        >>> ]
        >>> height, width = 16, 16
        >>> polys = PolygonMasks(masks, height, width)
        >>> nestshape(polys)

        >>> dc = BatchContainer([polys], stack=False)
        >>> print('dc = {}'.format(ub.repr2(dc, nl=1)))

        >>> num_masks, H, W = 3, 32, 32
        >>> rng = np.random.RandomState(0)
        >>> masks = (rng.rand(num_masks, H, W) > 0.1).astype(int)
        >>> bitmasks = BitmapMasks(masks, height=H, width=W)
        >>> nestshape(bitmasks)

        >>> dc = BatchContainer([bitmasks], stack=False)
        >>> print('dc = {}'.format(ub.repr2(dc, nl=1)))

    """

    def _recurse(d):
        if isinstance(d, dict):
            return ub.odict(sorted([(k, _recurse(v)) for k, v in d.items()]))

        clsname = type(d).__name__
        if 'Container' in clsname:
            meta = ub.odict(sorted([
                ('stack', d.stack),
                # ('padding_value', d.padding_value),
                # ('pad_dims', d.pad_dims),
                # ('datatype', d.datatype),
                ('cpu_only', d.cpu_only),
            ]))
            meta = ub.repr2(meta, nl=0)
            return {type(d).__name__ + meta: _recurse(d.data)}
        elif isinstance(d, list):
            return [_recurse(v) for v in d]
        elif isinstance(d, tuple):
            return tuple([_recurse(v) for v in d])
        elif isinstance(d, torch.Tensor):
            return d.shape
        elif isinstance(d, np.ndarray):
            return d.shape
        elif isinstance(d, (str, bytes)):
            return d
        elif isinstance(d, (int, float)):
            return d
        elif isinstance(d, slice):
            return d
        elif 'PolygonMasks' == clsname:
            # hack for mmdet
            return repr(d)
        elif 'BitmapMasks' == clsname:
            # hack for mmdet
            return repr(d)
        elif hasattr(d, 'shape'):
            return d.shape
        elif hasattr(d, 'items'):
            # hack for dict-like objects
            return ub.odict(sorted([(k, _recurse(v)) for k, v in d.items()]))
        elif d is None:
            return None
        else:
            raise TypeError(type(d))

    # globals()['_recurse'] = _recurse
    d = _recurse(data)
    return d


class BatchContainer(ub.NiceRepr):
    """
    A container for a set of items in a batch. Usually this is for network
    outputs or a set of items that have already been collated.

    Attributes:
        data (List[Any]): Unlike ItemContainer, data is always a list where
            len(data) is the number of devices this batch will run on.  Each
            item in the list may be either a pre-batched Tensor (in the case
            where the each item in the batch has the same shape) or a list of
            individual item Tensors (in the case where different batch items
            may have different shapes).
    """
    def __init__(self, data, stack=False, padding_value=-1, cpu_only=False,
                 pad_dims=2):
        self.data = data  # type: list
        self.meta = {
            'stack': stack,
            'padding_value': padding_value,
            'cpu_only': cpu_only,
            'pad_dims': pad_dims,
        }

    @property
    def nestshape(self):
        return nestshape(self.data)

    def numel(self):
        """
        The number of scalar elements held by this container
        """
        shapes = self.nestshape
        total = sum([np.prod(s) for s in shapes])
        return total

    @property
    def packshape(self):
        """
        The shape of this data if it was packed
        """
        # shape = np.maximum.reduce(self.nestshape)
        # return shape
        dim = 0
        if self.stack:
            # Should be a straight forward concatenation
            shapes = [d.shape for d in self.data]
            max_shape = np.maximum.reduce(shapes)  # should all be the same here
            stacked_dim = sum([s[dim] for s in shapes])
            max_shape[dim] = stacked_dim
            pack_shape = tuple(max_shape.tolist())
            return pack_shape
        else:
            shapes = nestshape(self.data)
            max_shape = np.maximum.reduce(shapes)
            stacked_dim = sum([s[dim] for s in shapes])
            max_shape[dim] = stacked_dim
            pack_shape = tuple(max_shape.tolist())
            return pack_shape

    def __nice__(self):
        try:
            shape_repr = ub.repr2(self.nestshape, nl=-2)
            return 'nestshape(data)={}'.format(shape_repr)
        except Exception:
            return object.__repr__(self)

    def __getitem__(self, index):
        cls = self.__class__
        return cls([d[index] for d in self.data], **self.meta)

    @property
    def cpu_only(self):
        return self.meta['cpu_only']

    @property
    def stack(self):
        return self.meta['stack']

    @property
    def padding_value(self):
        return self.meta['padding_value']

    @property
    def pad_dims(self):
        return self.meta['pad_dims']

    @classmethod
    def cat(cls, items, dim=0):
        """
        Concatenate data in multiple BatchContainers

        Example:
            >>> d1 = BatchContainer([torch.rand(3, 3, 1, 1), torch.rand(2, 3, 1, 1)])
            >>> d2 = BatchContainer([torch.rand(3, 1, 1, 1), torch.rand(2, 1, 1, 1)])
            >>> items = [d1, d2]
            >>> self = BatchContainer.cat(items, dim=1)
        """
        newdata = []
        num_devices = len(items[0].data)
        for device_idx in range(num_devices):
            parts = [item.data[device_idx] for item in items]
            newpart = torch.cat(parts, dim=dim)
            newdata.append(newpart)
        self = cls(newdata, **items[0].meta)
        return self

    @classmethod
    def demo(cls, key='img', n=5, num_devices=1):
        inbatch = [ItemContainer.demo(key) for _ in range(n)]
        self = ItemContainer._collate(inbatch, num_devices=num_devices)
        return self

    def pack(self):
        """
        Pack all of the data in this container into a single tensor.

        Returns:
            Tensor: packed data, padded with ``self.padding_value`` if
            ``self.stack`` is False.

        Example:
            >>> self = BatchContainer.demo('img')
            >>> print(self.pack())
            >>> self = BatchContainer.demo('box')
            >>> print(self.pack())
            >>> self = BatchContainer.demo('labels')
            >>> print(self.pack())
        """
        if self.stack:
            # Should be a straight forward concatenation
            packed = torch.cat(self.data, dim=0)
        else:
            # Need to account for padding values
            inbatch = list(ub.flatten(self.data))
            packed = padded_collate(inbatch, fill_value=self.padding_value)
        return packed

    def to(self, device):
        """ inplace move data onto a device """
        walker = ub.IndexableWalker(self.data)
        for path, val in walker:
            if torch.is_tensor(val):
                walker[path] = val.to(device)
        return self


default_collate = torch_data.dataloader.default_collate


def container_collate(inbatch, num_devices=None):
    """Puts each data field into a tensor/DataContainer with outer dimension
    batch size.

    Extend default_collate to add support for
    :type:`~mmcv.parallel.DataContainer`. There are 3 cases.

    1. cpu_only = True, e.g., meta data
    2. cpu_only = False, stack = True, e.g., images tensors
    3. cpu_only = False, stack = False, e.g., gt bboxes

    Example:
        >>> item1 = {
        >>>     'im': torch.rand(3, 512, 512),
        >>>     'label': torch.rand(3),
        >>> }
        >>> item2 = {
        >>>     'im': torch.rand(3, 512, 512),
        >>>     'label': torch.rand(3),
        >>> }
        >>> item3 = {
        >>>     'im': torch.rand(3, 512, 512),
        >>>     'label': torch.rand(3),
        >>> }
        >>> batch = batch_items = [item1, item2, item3]
        >>> raw_batch = container_collate(batch_items)
        >>> print('batch_items = {}'.format(ub.repr2(batch_items, nl=2)))
        >>> print('raw_batch = {}'.format(ub.repr2(raw_batch, nl=2)))

        >>> batch = batch_items = [
        >>>     {'im': ItemContainer.demo('img'), 'label': ItemContainer.demo('labels')},
        >>>     {'im': ItemContainer.demo('img'), 'label': ItemContainer.demo('labels')},
        >>>     {'im': ItemContainer.demo('img'), 'label': ItemContainer.demo('labels')},
        >>> ]
        >>> raw_batch = container_collate(batch, num_devices=2)
        >>> print('batch_items = {}'.format(ub.repr2(batch_items, nl=2)))
        >>> print('raw_batch = {}'.format(ub.repr2(raw_batch, nl=2)))

        >>> raw_batch = container_collate(batch, num_devices=6)
        >>> raw_batch = container_collate(batch, num_devices=3)
        >>> raw_batch = container_collate(batch, num_devices=4)
        >>> raw_batch = container_collate(batch, num_devices=1)
        >>> print('batch = {}'.format(ub.repr2(batch, nl=1)))
    """

    if not isinstance(inbatch, container_abcs.Sequence):
        raise TypeError("{} is not supported.".format(inbatch.dtype))
    item0 = inbatch[0]
    if isinstance(item0, ItemContainer):
        return item0.__class__._collate(inbatch, num_devices=num_devices)
    elif isinstance(item0, container_abcs.Sequence):
        transposed = zip(*inbatch)
        return [container_collate(samples,
                                  num_devices=num_devices)
                for samples in transposed]
    elif isinstance(item0, container_abcs.Mapping):
        return {
            key: container_collate([d[key] for d in inbatch],
                                   num_devices=num_devices)
            for key in item0
        }
    else:
        return default_collate(inbatch)
        # return _collate_else(inbatch, container_collate)


class PatchedRandomSampler(torch.utils.data.sampler.Sampler, ub.NiceRepr):
    r"""
    A modification of the standard torch Sampler that allows specification of
    ``num_samples``.

    See: https://github.com/pytorch/pytorch/pull/39214

    Example:
        >>> data_source = torch.arange(10)
        >>> # with replacement
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=None)
        >>> assert len(sampler) == 10 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=1)
        >>> assert len(sampler) == 1 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=5)
        >>> assert len(sampler) == 5 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=10)
        >>> assert len(sampler) == 10 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=True, num_samples=15)
        >>> assert len(sampler) == 15 == len(list(sampler))
        >>> # without replacement
        >>> sampler = PatchedRandomSampler(data_source, replacement=False, num_samples=None)
        >>> assert len(sampler) == 10 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=False, num_samples=1)
        >>> assert len(sampler) == 1 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=False, num_samples=5)
        >>> assert len(sampler) == 5 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=False, num_samples=10)
        >>> assert len(sampler) == 10 == len(list(sampler))
        >>> sampler = PatchedRandomSampler(data_source, replacement=False, num_samples=15)
        >>> assert len(sampler) == 10 == len(list(sampler))
    """

    def __init__(self, data_source, replacement=False, num_samples=None):
        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples

        if not isinstance(self.replacement, bool):
            raise ValueError("replacement should be a boolean value, but got "
                             "replacement={}".format(self.replacement))

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError("num_samples should be a positive integer "
                             "value, but got num_samples={}".format(self.num_samples))

    @property
    def num_samples(self):
        # dataset size might change at runtime
        if self._num_samples is None:
            num = len(self.data_source)
        else:
            if self.replacement:
                num = self._num_samples
            else:
                num = min(self._num_samples, len(self.data_source))
        return num

    def __iter__(self):
        n = len(self.data_source)
        if self.replacement:
            return iter(torch.randint(high=n, size=(self.num_samples,), dtype=torch.int64).tolist())
        return iter(torch.randperm(n).tolist()[:self.num_samples])

    def __len__(self):
        return self.num_samples


class SubsetSampler(torch.utils.data.sampler.Sampler, ub.NiceRepr):
    """
    Generates sample indices based on a specified order / subset

    Example:
        >>> indices = list(range(10))
        >>> assert indices == list(SubsetSampler(indices))
    """

    def __init__(self, indices):
        self.indices = indices

    def __iter__(self):
        for idx in self.indices:
            yield idx

    def __len__(self):
        return len(self.indices)


def padded_collate(inbatch, fill_value=-1):
    """
    Used for detection datasets with boxes.

    Example:
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 7
        >>> for i in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     n = 11 if i == 3 else rng.randint(0, 11)
        >>>     boxes = torch.rand(n, 4)
        >>>     item = (img, boxes)
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> assert list(out_batch[1].shape) == [bsize, 11, 4]

    Example:
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 4
        >>> for _ in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     #boxes = torch.empty(0, 4)
        >>>     boxes = torch.FloatTensor()
        >>>     item = (img, [boxes])
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> #assert list(out_batch[1][0].shape) == [bsize, 0, 4]
        >>> assert list(out_batch[1][0].shape) in [[0], []]  # torch .3 a .4

    Example:
        >>> inbatch = [torch.rand(4, 4), torch.rand(8, 4),
        >>>            torch.rand(0, 4), torch.rand(3, 4),
        >>>            torch.rand(0, 4), torch.rand(1, 4)]
        >>> out_batch = padded_collate(inbatch)
        >>> assert list(out_batch.shape) == [6, 8, 4]
    """
    try:
        if torch.is_tensor(inbatch[0]):
            num_items = [len(item) for item in inbatch]
            if ub.allsame(num_items):
                if len(num_items) == 0:
                    batch = torch.FloatTensor()
                elif num_items[0] == 0:
                    batch = torch.FloatTensor()
                else:
                    batch = default_collate(inbatch)
            else:
                max_size = max(num_items)
                real_tail_shape = None
                for item in inbatch:
                    if item.numel():
                        tail_shape = item.shape[1:]
                        if real_tail_shape is not None:
                            assert real_tail_shape == tail_shape
                        real_tail_shape = tail_shape

                padded_inbatch = []
                for item in inbatch:
                    n_extra = max_size - len(item)
                    if n_extra > 0:
                        shape = (n_extra,) + tuple(real_tail_shape)
                        if torch.__version__.startswith('0.3'):
                            extra = torch.Tensor(np.full(shape, fill_value=fill_value))
                        else:
                            extra = torch.full(shape, fill_value=fill_value,
                                               dtype=item.dtype)
                        padded_item = torch.cat([item, extra], dim=0)
                        padded_inbatch.append(padded_item)
                    else:
                        padded_inbatch.append(item)
                batch = inbatch
                batch = default_collate(padded_inbatch)
        else:
            batch = _collate_else(inbatch, padded_collate)
    except Exception as ex:
        if not isinstance(ex, CollateException):
            try:
                _debug_inbatch_shapes(inbatch)
            except Exception:
                pass
            raise CollateException(
                'Failed to collate inbatch={}. Reason: {!r}'.format(inbatch, ex))
        else:
            raise
    return batch


default_collate = torch_data.dataloader.default_collate


# numpy_type_map = torch_data.dataloader.numpy_type_map  # moved in torch 1.1.0
numpy_type_map = {
    'float64': torch.DoubleTensor,
    'float32': torch.FloatTensor,
    'float16': torch.HalfTensor,
    'int64': torch.LongTensor,
    'int32': torch.IntTensor,
    'int16': torch.ShortTensor,
    'int8': torch.CharTensor,
    'uint8': torch.ByteTensor,
}


class CollateException(Exception):
    pass


_DEBUG = False


def _collate_else(batch, collate_func):
    """
    Handles recursion in the else case for these special collate functions

    This is duplicates all non-tensor cases from `torch_data.dataloader.default_collate`
    This also contains support for collating slices.
    """
    error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
    elem_type = type(batch[0])
    if elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if re.search('[SaUO]', elem.dtype.str) is not None:
                raise TypeError(error_msg.format(elem.dtype))

            return torch.stack([torch.from_numpy(b) for b in batch], 0)
        if elem.shape == ():  # scalars
            py_type = float if elem.dtype.name.startswith('float') else int
            return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
    elif isinstance(batch[0], slice):
        batch = default_collate([{
            'start': sl.start,
            'stop': sl.stop,
            'step': 1 if sl.step is None else sl.step
        } for sl in batch])
        return batch
    elif isinstance(batch[0], int_classes):
        return torch.LongTensor(batch)
    elif isinstance(batch[0], float):
        return torch.DoubleTensor(batch)
    elif isinstance(batch[0], string_classes):
        return batch
    elif isinstance(batch[0], container_abcs.Mapping):
        # Hack the mapping collation implementation to print error info
        if _DEBUG:
            collated = {}
            try:
                for key in batch[0]:
                    collated[key] = collate_func([d[key] for d in batch])
            except Exception:
                print('\n!!Error collating key = {!r}\n'.format(key))
                raise
            return collated
        else:
            return {key: collate_func([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], tuple) and hasattr(batch[0], '_fields'):  # namedtuple
        return type(batch[0])(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(batch[0], container_abcs.Sequence):
        transposed = zip(*batch)
        return [collate_func(samples) for samples in transposed]
    else:
        raise TypeError((error_msg.format(type(batch[0]))))


def _debug_inbatch_shapes(inbatch):
    import ubelt as ub
    print('len(inbatch) = {}'.format(len(inbatch)))
    extensions = ub.util_format.FormatterExtensions()
    #
    @extensions.register((torch.Tensor, np.ndarray))
    def format_shape(data, **kwargs):
        return ub.repr2(dict(type=str(type(data)), shape=data.shape), nl=1, sv=1)
    print('inbatch = ' + ub.repr2(inbatch, extensions=extensions, nl=True))



from reformer_pytorch import LSHSelfAttention
import ubelt as ub

try:
    import xdev
    profile = xdev.profile
except Exception:
    profile = ub.identity


class ReformerMultiheadedSelfAttention(LSHSelfAttention):
    """
    This seems like a good idea, but either I'm using it wrong or the
    C-bindings in normal attention make this lose all of its benefit.

    Ignore:
        from geowatch.tasks.fusion.architectures.transformer import *  # NOQA
        D = 9  # embedding dimension
        H = 3   # number of heads
        B = 5   # batch size
        S = 7   # sequence length
        x = torch.rand(S, B, D)

        self = ReformerMultiheadedSelfAttention(D, H)

        MultiheadSelfAttention(D, H)(x).shape
        ReformerMultiheadedSelfAttention(D, H)(x)
        from reformer_pytorch import LSHAttention
        q = einops.rearrange(x, 's b (h e) -> b h s e', h=H)
        FastAttention(dim_heads=D // H, nb_features=None)(q, q, q).shape
    """

    def __init__(self, embed_dim, num_heads):
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        assert embed_dim % num_heads == 0
        dim_heads = embed_dim // num_heads
        self.dim_heads = dim_heads
        # nb_features = int(dim_heads * math.log(dim_heads))
        # nb_features = int(dim_heads * 2)
        super().__init__(
            dim=embed_dim, heads=num_heads, dim_head=dim_heads,
            bucket_size=64, n_hashes=8, causal=False)

    @profile
    def forward(self, x, key_padding_mask=None):
        if key_padding_mask is not None:
            raise NotImplementedError
        s, b, he = x.shape
        bsd = x.permute(1, 0, 2)
        # a = LSHSelfAttention.forward(self, bsd)
        a = super().forward(bsd)
        out = a.permute(1, 0, 2)
        return out



from typing import Tuple, Union
import torch

from kornia.geometry.transform import vflip, rotate

UnionType = Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]


def random_rotate(input: torch.Tensor) -> UnionType:
    r"""Rotate a tensor image or a batch of tensor images randomly.
    Input should be a tensor of shape (C, H, W) or a batch of tensors :math:`(*, C, H, W)`.
    Args:
        input tensor.
    Returns:
        torch.Tensor: The rotated input
    """

    if not torch.is_tensor(input):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(input)}")

    device: torch.device = input.device
    input = input.unsqueeze(0)
    input = input.view(-1, *input.shape[-3:])
    angle: torch.Tensor = torch.empty(input.shape[0], device=device).uniform_(-180, -180)

    rotated = rotate(input, angle)
    return rotated


def random_vflip(input: torch.Tensor, p: float = 0.5, return_transform: bool = False) -> UnionType:
    r"""Vertically flip a tensor image or a batch of tensor images randomly with a given probability.
    Input should be a tensor of shape (C, H, W) or a batch of tensors :math:`(*, C, H, W)`.
    Args:
        p (float): probability of the image being flipped. Default value is 0.5
        return_transform (bool): if ``True`` return the matrix describing the transformation applied to each
        input tensor.
    Returns:
        torch.Tensor: The vertically flipped input
        torch.Tensor: The applied transformation matrix :math: `(*, 3, 3)` if return_transform flag
                      is set to ``True``
    """

    if not torch.is_tensor(input):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(input)}")

    if not isinstance(p, float):
        raise TypeError(f"The probability should be a float number. Got {type(p)}")

    if not isinstance(return_transform, bool):
        raise TypeError(f"The return_transform flag must be a bool. Got {type(return_transform)}")

    device: torch.device = input.device
    dtype: torch.dtype = input.dtype

    input = input.unsqueeze(0)
    input = input.view(-1, *input.shape[-3:])

    probs: torch.Tensor = torch.empty(input.shape[0], device=device).uniform_(0, 1)

    to_flip: torch.Tensor = probs < p
    flipped: torch.Tensor = input.clone()

    flipped[to_flip] = vflip(input[to_flip])

    if return_transform:

        trans_mat: torch.Tensor = torch.eye(3, device=device, dtype=dtype).expand(input.shape[0], -1, -1)

        w: int = input.shape[-2]
        flip_mat: torch.Tensor = torch.tensor([[-1, 0, w],
                                               [0, 1, 0],
                                               [0, 0, 1]])

        trans_mat[to_flip] = flip_mat.to(device).to(dtype)

        return flipped, trans_mat

    return flipped



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the DeepLab meta architecture."""
import collections
import functools
from typing import Any, Dict, Text, Tuple

import tensorflow as tf
from absl import logging
from deeplab2 import common
from deeplab2 import config_pb2
# from deeplab2.data import dataset
from deeplab2.model import builder
from deeplab2.model import utils

# from deeplab2.model.post_processor import post_processor_builder

_OFFSET_OUTPUT = 'offset'


class DeepLab(tf.keras.Model):
    """This class represents the DeepLab meta architecture.

    This class supports four architectures of the DeepLab family: DeepLab V3,
    DeepLab V3+, Panoptic-DeepLab, and MaX-DeepLab. The exact architecture must be
    defined during initialization.
    """

    def __init__(self,
                 config: config_pb2.ExperimentOptions, use_ln=False):
        # dataset_descriptor: dataset.DatasetDescriptor):
        """Initializes a DeepLab architecture.

        Args:
          config: A config_pb2.ExperimentOptions configuration.
          dataset_descriptor: A dataset.DatasetDescriptor.

        Raises:
          ValueError: If MaX-DeepLab is used with multi-scale inference.
        """
        super(DeepLab, self).__init__(name='DeepLab')

        if config.trainer_options.solver_options.use_sync_batchnorm:
            logging.info('Synchronized Batchnorm is used.')
            bn_layer = functools.partial(
                tf.keras.layers.experimental.SyncBatchNormalization,
                momentum=config.trainer_options.solver_options.batchnorm_momentum,
                epsilon=config.trainer_options.solver_options.batchnorm_epsilon)
        else:
            logging.info('Standard (unsynchronized) Batchnorm is used.')
            bn_layer = functools.partial(
                tf.keras.layers.BatchNormalization,
                momentum=config.trainer_options.solver_options.batchnorm_momentum,
                epsilon=config.trainer_options.solver_options.batchnorm_epsilon)

        if use_ln:
            bn_layer = tf.keras.layers.LayerNormalization

        # Divide weight decay by 2 to match the implementation of tf.nn.l2_loss. In
        # this way, we allow our users to use a normal weight decay (e.g., 1e-4 for
        # ResNet variants) in the config textproto. Then, we pass the adjusted
        # weight decay (e.g., 5e-5 for ResNets) to keras in order to exactly match
        # the commonly used tf.nn.l2_loss in TF1. References:
        # https://github.com/tensorflow/models/blob/68ee72ae785274156b9e943df4145b257cd78b32/official/vision/beta/tasks/image_classification.py#L41
        # https://www.tensorflow.org/api_docs/python/tf/keras/regularizers/l2
        # https://www.tensorflow.org/api_docs/python/tf/nn/l2_loss
        self._encoder = builder.create_encoder(
            config.model_options.backbone, bn_layer,
            conv_kernel_weight_decay=(
                    config.trainer_options.solver_options.weight_decay / 2))

        self._decoder = builder.create_decoder(
            config.model_options, bn_layer, -1e10)  # dataset_descriptor.ignore_label)

        self._is_max_deeplab = (
                config.model_options.WhichOneof('meta_architecture') == 'max_deeplab')
        self._post_processor = None  # post_processor_builder.get_post_processor(
        # config, dataset_descriptor)

        # The ASPP pooling size is always set to train crop size, which is found to
        # be experimentally better.
        pool_size = config.train_dataset_options.crop_size
        output_stride = float(config.model_options.backbone.output_stride)
        pool_size = tuple(
            utils.scale_mutable_sequence(pool_size, 1.0 / output_stride))
        logging.info('Setting pooling size to %s', pool_size)
        self.set_pool_size(pool_size)

        # Variables for multi-scale inference.
        self._add_flipped_images = config.evaluator_options.add_flipped_images
        if not config.evaluator_options.eval_scales:
            self._eval_scales = [1.0]
        else:
            self._eval_scales = config.evaluator_options.eval_scales
        if self._is_max_deeplab and (
                self._add_flipped_images or len(self._eval_scales) > 1):
            raise ValueError(
                'MaX-DeepLab does not support multi-scale inference yet.')

    def call(self,
             input_tensor: tf.Tensor,
             training: bool = False) -> Dict[Text, Any]:
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels]. The input tensor should contain batches of RGB images.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the results of the specified DeepLab architecture.
          The results are bilinearly upsampled to input size before returning.
        """
        # Normalize the input in the same way as Inception. We normalize it outside
        # the encoder so that we can extend encoders to different backbones without
        # copying the normalization to each encoder. We normalize it after data
        # preprocessing because it is faster on TPUs than on host CPUs. The
        # normalization should not increase TPU memory consumption because it does
        # not require gradient.
        input_tensor = input_tensor / 127.5 - 1.0
        # Get the static spatial shape of the input tensor.
        _, input_h, input_w, _ = input_tensor.get_shape().as_list()

        if training:

            result_dict = self._decoder(
                self._encoder(input_tensor, training=training), training=training)
            result_dict = self._resize_predictions(
                result_dict,
                target_h=input_h,
                target_w=input_w)
        else:

            result_dict = collections.defaultdict(list)
            # Evaluation mode where one could perform multi-scale inference.
            scale_1_pool_size = self.get_pool_size()
            logging.info('Eval with scales %s', self._eval_scales)
            for eval_scale in self._eval_scales:
                # Get the scaled images/pool_size for each scale.
                scaled_images, scaled_pool_size = (
                    self._scale_images_and_pool_size(
                        input_tensor, list(scale_1_pool_size), eval_scale))
                # Update the ASPP pool size for different eval scales.
                self.set_pool_size(tuple(scaled_pool_size))
                logging.info('Eval scale %s; setting pooling size to %s',
                             eval_scale, scaled_pool_size)
                pred_dict = self._decoder(
                    self._encoder(scaled_images, training=training), training=training)
                # MaX-DeepLab skips this resizing and upsamples the mask outputs in
                # self._post_processor.
                pred_dict = self._resize_predictions(
                    pred_dict,
                    target_h=input_h,
                    target_w=input_w)
                # Change the semantic logits to probabilities with softmax. Note
                # one should remove semantic logits for faster inference. We still
                # keep them since they will be used to compute evaluation loss.
                pred_dict[common.PRED_SEMANTIC_PROBS_KEY] = tf.nn.softmax(
                    pred_dict[common.PRED_SEMANTIC_LOGITS_KEY])
                # Store the predictions from each scale.
                for output_type, output_value in pred_dict.items():
                    result_dict[output_type].append(output_value)
                if self._add_flipped_images:
                    pred_dict_reverse = self._decoder(
                        self._encoder(tf.reverse(scaled_images, [2]), training=training),
                        training=training)
                    pred_dict_reverse = self._resize_predictions(
                        pred_dict_reverse,
                        target_h=input_h,
                        target_w=input_w,
                        reverse=True)
                    # Change the semantic logits to probabilities with softmax.
                    pred_dict_reverse[common.PRED_SEMANTIC_PROBS_KEY] = tf.nn.softmax(
                        pred_dict_reverse[common.PRED_SEMANTIC_LOGITS_KEY])
                    # Store the predictions from each scale.
                    for output_type, output_value in pred_dict_reverse.items():
                        result_dict[output_type].append(output_value)
            # Set back the pool_size for scale 1.0, the original setting.
            self.set_pool_size(tuple(scale_1_pool_size))
            # Average results across scales.
            for output_type, output_value in result_dict.items():
                result_dict[output_type] = tf.reduce_mean(
                    tf.stack(output_value, axis=0), axis=0)
            # Post-process the results.
            # result_dict.update(self._post_processor(result_dict))

        if common.PRED_CENTER_HEATMAP_KEY in result_dict:
            result_dict[common.PRED_CENTER_HEATMAP_KEY] = tf.squeeze(
                result_dict[common.PRED_CENTER_HEATMAP_KEY], axis=3)
        return result_dict

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layer to global average pooling."""
        self._decoder.reset_pooling_layer()

    def set_pool_size(self, pool_size: Tuple[int, int]):
        """Sets the pooling size of the ASPP pooling layer.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layer.
        """
        self._decoder.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._decoder.get_pool_size()

    @property
    def checkpoint_items(self) -> Dict[Text, Any]:
        items = dict(encoder=self._encoder)
        items.update(self._decoder.checkpoint_items)
        return items

    @property
    def auxiliary_output_number(self) -> int:
        # auxiliary_output_number is only supported in K-MaX meta, thus we hard
        # code it to 0 here.
        return 0

    def _resize_predictions(self, result_dict, target_h, target_w, reverse=False):
        """Resizes predictions to the target height and width.

        This function resizes the items in the result_dict to the target height and
        width. The items are optionally reversed w.r.t width if `reverse` is True.

        Args:
          result_dict: A dictionary storing prediction results to be resized.
          target_h: An integer, the target height.
          target_w: An integer, the target width.
          reverse: A boolean, reversing the prediction result w.r.t. width.

        Returns:
          Resized (or optionally reversed) result_dict.
        """
        # The default MaX-DeepLab paper does not upsample any output during training
        # in order to save GPU/TPU memory, but upsampling might lead to better
        # performance.
        if self._is_max_deeplab:
            return result_dict
        for key, value in result_dict.items():
            if reverse:
                value = tf.reverse(value, [2])
                # Special care to offsets: need to flip x-offsets.
                if _OFFSET_OUTPUT in key:
                    offset_y, offset_x = tf.split(
                        value=value, num_or_size_splits=2, axis=3)
                    offset_x *= -1
                    value = tf.concat([offset_y, offset_x], 3)
            if _OFFSET_OUTPUT in key:
                result_dict[key] = utils.resize_and_rescale_offsets(
                    value, [target_h, target_w])
            else:
                result_dict[key] = utils.resize_bilinear(
                    value, [target_h, target_w])
        return result_dict

    def _scale_images_and_pool_size(self, images, pool_size, scale):
        """Scales images and pool_size w.r.t. scale.

        Args:
          images: An input tensor with shape [batch, height, width, 3].
          pool_size: A list with two elements, specifying the pooling size
            of ASPP pooling layer.
          scale: A float, used to scale the input images and pool_size.

        Returns:
          Scaled images, and pool_size.
        """
        if scale == 1.0:
            scaled_images = images
            scaled_pool_size = pool_size
        else:
            image_size = images.get_shape().as_list()[1:3]
            scaled_image_size = utils.scale_mutable_sequence(image_size, scale)
            scaled_images = utils.resize_bilinear(images, scaled_image_size)
            scaled_pool_size = [None, None]
            if pool_size != [None, None]:
                scaled_pool_size = utils.scale_mutable_sequence(pool_size, scale)
        return scaled_images, scaled_pool_size



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from argparse import Namespace
import pytorch_lightning as pl
from torchmetrics.classification.accuracy import Accuracy

from ..data.datasets import kwcoco_dataset
from ..utils.unet_blur import UNetEncoder, UNetDecoder
from ..utils.focal_loss import BinaryFocalLoss


class pretext(pl.LightningModule):
    TASK_NAMES = [
        'sort',
        'augment',
        'overlap'
    ]

    def __init__(self, hparams):
        super().__init__()

        if isinstance(hparams, dict):
            hparams = Namespace(**hparams)

        self.save_hyperparameters(hparams)

        if hparams.train_dataset is not None:
            self.trainset = kwcoco_dataset(hparams.train_dataset, hparams.sensor, hparams.bands, hparams.patch_size)
        else:
            self.trainset = None
        if hparams.vali_dataset is not None:
            self.valset = kwcoco_dataset(hparams.vali_dataset, hparams.sensor, hparams.bands, hparams.patch_size)
        else:
            self.valset = None

        print('hparams = {!r}'.format(hparams))
        if hasattr(hparams, 'num_channels'):
            # hack for loading without dataset state
            num_channels = hparams.num_channels
        elif self.trainset is not None:
            num_channels = self.trainset.num_channels()
        else:
            num_channels = 6

        # determine which tasks to run
        self.task_indices = []
        # no tasks specified
        if len(self.hparams.tasks) < 1:
            raise ValueError(f'tasks must be specified. Options are {", ".join(pretext.TASK_NAMES)}, or all')
        # perform all tasks
        elif len(self.hparams.tasks) == 1 and self.hparams.tasks[0].lower() == 'all':
            self.task_indices = [i for i in range(len(pretext.TASK_NAMES))]
        # run a subset of tasks
        else:
            for task in self.hparams.tasks:
                if task.lower() in pretext.TASK_NAMES:
                    self.task_indices.append(pretext.TASK_NAMES.index(task.lower()))
                else:
                    raise ValueError(f'\'{task}\' not recognized as an available task. Options are {", ".join(pretext.TASK_NAMES)}, or all')
        if pretext.TASK_NAMES.index('augment') in self.task_indices and pretext.TASK_NAMES.index('overlap') not in self.task_indices:
            self.task_indices.append(pretext.TASK_NAMES.index('overlap'))

        # shared model body
        self.encoder = UNetEncoder(in_channels=num_channels)
        self.decoder = UNetDecoder(out_channels=self.hparams.feature_dim_shared)

        # task specific necks
        self.necks = [
            self.task_neck(self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # sort task
            self.task_neck(self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # augment task
            self.task_neck(self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # overlap task
        ]
        self.necks = nn.ModuleList([ self.necks[i] for i in self.task_indices ])

        # task specific heads
        self.heads = [
            self.pixel_classification_head(2 * self.hparams.feature_dim_each_task),  # sort task
            self.image_classification_head( self.hparams.feature_dim_each_task),  # augment task
            self.image_classification_head( self.hparams.feature_dim_each_task),  # overlap task
        ]
        self.heads = nn.ModuleList([ self.heads[i] for i in self.task_indices ])
        # task specific criterion
        self.criteria = [
            BinaryFocalLoss(gamma=self.hparams.focal_gamma),  # sort task
            nn.TripletMarginLoss(),  # augment task
            nn.TripletMarginLoss(),  # overlap task
        ]
        self.criteria = [ self.criteria[i] for i in self.task_indices ]

        # task specific metrics
        self.sort_accuracy = Accuracy()

    def forward(self, image):
        # pass through shared model body
        encoded = self.encoder(image)
        decoded = self.decoder(encoded)
        return decoded

    def shared_step(self, batch):
        # get features of each image from shared model body
        image1_features = self(batch['image1'])
        image2_features = self(batch['image2'])
        offset_image1_features = self(batch['offset_image1'])
        augmented_image1_features = self(batch['augmented_image1'])
        # get time sort labels
        time_sort_labels = batch['time_sort_label']
        time_sort_labels = time_sort_labels.unsqueeze(1).unsqueeze(1).repeat(1, self.hparams.patch_size, self.hparams.patch_size).unsqueeze(1)

        losses = []
        output = {}

        # Time Sort task
        if 0 in self.task_indices:
            module_list_idx = self.task_indices.index(0)
            # forward pass through neck
            image1_sort_out = self.necks[module_list_idx](image1_features)
            image2_sort_out = self.necks[module_list_idx](image2_features)
            # forward pass through head
            time_sort_prediction = self.heads[module_list_idx](torch.cat((image1_sort_out, image2_sort_out), dim=1))
            # evaluate
            loss_time = self.criteria[module_list_idx](time_sort_prediction, time_sort_labels)
            if self.hparams.aot_penalty_weight:
                l1_penalty = torch.norm(image1_sort_out - image2_sort_out, 1, dim=1) / image1_sort_out.shape[1]
                l1_penalty_filtered = -1 * torch.topk(-1 * l1_penalty.flatten(), int(self.hparams.aot_penalty_percentage * l1_penalty.numel())).values
                loss_time = loss_time + self.hparams.aot_penalty_weight * l1_penalty_filtered.mean()
            time_accuracy = self.sort_accuracy((time_sort_prediction > 0.), time_sort_labels.int())

            losses.append(loss_time)
            output['time_accuracy'] = time_accuracy
            output['loss_time_sort'] = loss_time.detach()
            output['before_after_heatmap'] = F.sigmoid(time_sort_prediction)
        # Overlap task
        if 2 in self.task_indices:
            module_list_idx = self.task_indices.index(2)
            # forward pass through neck
            image1_overlap_out = self.necks[module_list_idx](image1_features)
            image2_overlap_out = self.necks[module_list_idx](image2_features)
            image1_offset_overlap_out = self.necks[module_list_idx](offset_image1_features)
            # forward pass through head
            image1_overlap_out = self.heads[module_list_idx](image1_overlap_out)
            image2_overlap_out = self.heads[module_list_idx](image2_overlap_out)
            image1_offset_overlap_out = self.heads[module_list_idx](image1_offset_overlap_out)
            # evaluate
            loss_offset = self.criteria[module_list_idx](image1_overlap_out, image2_overlap_out, image1_offset_overlap_out)

            losses.append(loss_offset)
            if 2 in self.task_indices:
                output['loss_offset'] = loss_offset.detach()

        # Augment task
        if 1 in self.task_indices:
            module_list_idx = self.task_indices.index(1)
            # image1 forward pass through neck
            image1_augment_out = self.necks[module_list_idx](image1_features)
            image1_augmented_augment_out = self.necks[module_list_idx](augmented_image1_features)
            # image1 forward pass through head
            image1_augment_out = self.heads[module_list_idx](image1_augment_out)
            image1_augmented_augment_out = self.heads[module_list_idx](image1_augmented_augment_out)
            # image1 evaluate
            loss_augment = self.criteria[module_list_idx](image1_augment_out, image1_augmented_augment_out, image1_offset_overlap_out)

            losses.append(loss_augment)
            output['loss_augmented'] = loss_augment.detach()

        # add up loss
        loss = torch.sum(torch.stack(losses))
        output['loss'] = loss

        return output

    def training_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        self.log_dict(output)
        return output

    def validation_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        output = {"val_" + key: val for key, val in output.items()}
        self.log_dict(output)
        return output

    def predict(self, batch):
        self.eval()
        feats = {}
        shared_feats = self(batch)
        feats['shared'] = shared_feats
        return feats['shared']

    ### Temporary solution
    def predict_before_after(self, image):
        self.eval()
        im1 = image[:, 0, :]
        im2 = image[:, 1, :]
        image1_features = self(im1)
        image2_features = self(im2)
        image1_sort_out = self.necks[self.task_indices.index(0)](image1_features)
        image2_sort_out = self.necks[self.task_indices.index(0)](image2_features)
        # forward pass through head
        time_sort_prediction = self.heads[self.task_indices.index(0)](torch.cat((image1_sort_out, image2_sort_out), dim=1))
        return time_sort_prediction

    def train_dataloader(self):
        return DataLoader(self.trainset, batch_size=self.hparams.batch_size, num_workers=self.hparams.workers, shuffle=True)

    def val_dataloader(self):
        if self.hparams.vali_dataset is not None:
            return DataLoader(self.valset, batch_size=self.hparams.batch_size, num_workers=self.hparams.workers)

    def configure_optimizers(self):
        opt = torch.optim.AdamW(
            self.parameters(),
            lr=self.hparams.learning_rate)

        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            opt, step_size=self.hparams.step_size,
            gamma=self.hparams.lr_gamma)

        return {'optimizer': opt, 'lr_scheduler': lr_scheduler}

    def task_neck(self, in_chan, out_chan):
        kernel_size = 1
        stride = 1
        padding = int((kernel_size - 1) / 2)
        return nn.Sequential(
            nn.Conv2d(in_chan, in_chan, kernel_size, stride, padding),
            nn.BatchNorm2d(in_chan),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_chan, out_chan, kernel_size, stride, padding),
            nn.BatchNorm2d(out_chan),
            nn.ReLU(inplace=True)
        )

    def pixel_classification_head(self, in_chan):
        kernel_size = 1
        stride = 1
        padding = int((kernel_size - 1) / 2)
        return nn.Sequential(
                            nn.Conv2d(in_chan, in_chan, kernel_size, stride, padding),
                            nn.BatchNorm2d(in_chan),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(in_chan, 1, 1, bias=False, padding=0),
                            )

    def image_classification_head(self, in_chan):
        return nn.Sequential(
                             nn.AdaptiveAvgPool2d(output_size=(1, 1)),
                             nn.Flatten(1, -1),
                             nn.Linear(in_chan, 2),
                            )



#!/usr/bin/env python
import sys
import subprocess
import json
import traceback
import requests
import ubelt as ub
import scriptconfig as scfg


class UploadRGDConfig(scfg.DataConfig):
    """
    Run TA-2 BAS fusion as baseline framework component
    """
    input_site_models_s3 = scfg.Value(None, type=str, position=1, help='Path to S3 directory of site models')

    rgd_aws_region = scfg.Value(None, type=str, help='AWS region where RGD instance is running')

    rgd_deployment_name = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Name of RGD deployment (e.g. 'resonantgeodatablue'
        '''))

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    title = scfg.Value(None, type=str, help='Title of the model run')

    region_id = scfg.Value(None, type=str, help='Region ID (e.g. "KR_R002")')

    performer_shortcode = scfg.Value('KIT', type=str, help='Performer shortcode (e.g. "KIT")')

    rgd_endpoint_override = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Use this RGD URL instead of looking up via aws tools
        '''))

    jobs = scfg.Value(8, type=int, short_alias=['j'], help='Number of jobs to run in parallel')

    expiration_time = scfg.Value(None, type=int, short_alias=['x'], help=ub.paragraph(
        '''
        Number of days to keep system run output in RGD
        '''))


def main(cmdline=1, **kwargs):
    config = UploadRGDConfig.cli(cmdline=cmdline, data=kwargs)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    assert config.rgd_aws_region is not None
    assert config.input_site_models_s3 is not None
    assert config.rgd_deployment_name is not None
    assert config.title is not None
    assert config.region_id is not None
    upload_to_rgd(config)


def get_model_results(model_run_results_url):
    model_runs_result = requests.get(model_run_results_url, params={
        'limit': '0'})
    request_json = model_runs_result.json()
    request_results = request_json.get('results', ())
    return request_results


def upload_to_rgd(config):

    input_site_models_s3 = config.input_site_models_s3
    rgd_aws_region = config.rgd_aws_region
    rgd_deployment_name = config.rgd_deployment_name
    title = config.title
    region_id = config.region_id
    aws_profile = config.aws_profile
    performer_shortcode = config.performer_shortcode
    jobs = config.jobs
    rgd_endpoint_override = config.rgd_endpoint_override
    expiration_time = config.expiration_time

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # Ensure performer_shortcode is uppercase
    performer_shortcode = performer_shortcode.upper()

    if aws_profile is not None:
        aws_base_command = ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    local_site_models_dir = '/tmp/site_models'
    ub.cmd([*aws_base_command, '--recursive',
            input_site_models_s3, local_site_models_dir],
           check=True, verbose=3)

    if rgd_endpoint_override is None:
        try:
            endpoint_result = ub.cmd(
                ['aws',
                 *(('--profile', aws_profile)
                   if aws_profile is not None else ()),
                 '--region', rgd_aws_region,
                 'elbv2', 'describe-load-balancers',
                 '--region', rgd_aws_region,
                 '--names', "{}-internal-alb".format(rgd_deployment_name)],
                check=True,
                verbose=3)
        except subprocess.CalledProcessError as e:
            print(e.stderr, file=sys.stderr)
            raise e

        rgd_instance_details = json.loads(endpoint_result.stdout)
        rgd_endpoint = rgd_instance_details['LoadBalancers'][0]['DNSName']
    else:
        rgd_endpoint = rgd_endpoint_override

    # Check that our run doesn't already exist
    model_run_results_url = f"http://{rgd_endpoint}/api/model-runs/"

    # from retry.api import retry_call
    from geowatch.utils.util_retry import retry_call
    from geowatch.utils import util_framework
    logger = util_framework.PrintLogger()
    print('\nAttempt to post to RGD model results endpoint')
    print(f'model_run_results_url = {ub.urepr(model_run_results_url, nl=1)}')

    request_results = retry_call(
        get_model_results, fargs=[model_run_results_url], tries=3,
        exceptions=(Exception,), delay=3, logger=logger)

    print('Got Results')
    existing_model_run = None

    for model_run in request_results:
        if (model_run['title'] != title or
             model_run['performer']['short_code'] != performer_shortcode):
            continue

        model_region = model_run.get('region')
        if model_region is None:
            continue

        if (isinstance(model_region, dict) and
             model_run['region'].get('name') == region_id):
            existing_model_run = model_run
            break
        elif (isinstance(model_region, str) and
               model_region == region_id):  # noqa
            existing_model_run = model_run
            break

    if existing_model_run is not None:
        model_run_id = model_run['id']
    else:
        post_model_url = f"http://{rgd_endpoint}/api/model-runs/"
        post_model_data = {"performer": performer_shortcode,
                           "title": title,
                           "region": {"name": region_id},
                           "parameters": {}}

        if expiration_time is not None:
            post_model_data['expiration_time'] = expiration_time

        post_model_result = requests.post(
            post_model_url,
            json=post_model_data,
            headers={"Content-Type": "application/json"})

        model_run_id = post_model_result.json()['id']

    post_site_url = (
        f"http://{rgd_endpoint}/api/model-runs/{model_run_id}/site-model/")

    local_site_models_dir = ub.Path(local_site_models_dir)

    jobs = ub.JobPool(mode='process', max_workers=jobs)
    for site_filepath in local_site_models_dir.glob('*.geojson'):
        jobs.submit(post_site, post_site_url, site_filepath)

    for site_post_job in ub.ProgIter(jobs.as_completed(),
                                     total=len(jobs),
                                     desc='Uploading sites..'):
        try:
            result = site_post_job.result()
        except Exception:
            print("Exception occurred (printed below)")
            traceback.print_exception(*sys.exc_info())
            continue
        else:
            if result.status_code != 201:
                print(f"Error uploading site, status "
                      f"code: [{result.status_code}]")
                print(result.text)


def post_site(post_site_url, site_filepath):
    with open(site_filepath, 'r') as f:
        response = requests.post(
            post_site_url,
            json=json.load(f),
            headers={"Content-Type": "application/json"})

    return response


if __name__ == "__main__":
    main()



"""
Compute intensity histograms of the underlying images in this dataset.

TODO:

    - [ ] Migrate to kwcoco proper

    - [ ] Fix accumulation of floating point values

    - [ ] Handle nodata

    - [x] Rename to coco_spectra

CommandLine:
    geowatch spectra geowatch-msi --dst foo.png
    geowatch spectra --src special:geowatch-msi --show=True --stat=density
    geowatch spectra --src special:photos --show=True --fill=False
    geowatch spectra --src special:shapes8 --show=True --stat=count --cumulative=True --multiple=stack

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    geowatch spectra --src $DVC_DATA_DPATH/Drop6/data.kwcoco.zip --channels='red|green|blue|nir' --workers=11

    DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc
    KWCOCO_FPATH=$DVC_DPATH/Drop2-Aligned-TA1-2022-01/combo_L_nowv_vali.kwcoco.json
    geowatch spectra --src $KWCOCO_FPATH --show=True --show=True --include_channels="forest|water|bare_ground"
"""
import ubelt as ub
import scriptconfig as scfg
import math


class CocoSpectraConfig(scfg.DataConfig):
    """
    Plot the spectrum of band intensities in a kwcoco file.
    """
    __default__ = {
        'src': scfg.Value('data.kwcoco.json', help='input coco dataset', position=1),

        'dst': scfg.Value(None, help='if specified dump the figure to disk at this file path (e.g. with a jpg or png suffix)'),

        'cache_dpath': scfg.Value(None, help='if specified, read/write cached stats files from here instead of as sidecar files'),

        'show': scfg.Value(False, isflag=True, help='if True, do a plt.show()'),
        'draw': scfg.Value(True, isflag=True, help='if False disables all visualization and just print tables'),

        'workers': scfg.Value(0, help='number of io workers'),
        'mode': scfg.Value('process', help='type of parallelism'),

        'include_channels': scfg.Value(None, help='if specified can be | separated valid channels', alias=['channels']),
        'exclude_channels': scfg.Value(None, help='if specified can be | separated invalid channels'),

        'include_sensors': scfg.Value(None, help='if specified can be comma separated valid sensors'),
        'exclude_sensors': scfg.Value(None, help='if specified can be comma separated invalid sensors'),

        'valid_range': scfg.Value(None, help='Only include values within this range; specified as <min_val>:<max_val> e.g. (0:10000)'),

        'title': scfg.Value(None, help='Provide a title for the histogram figure'),

        'max_images': scfg.Value(None, help='if given only sample this many images when computing statistics', group='subset'),

        'select_images': scfg.Value(
            None, type=str, help=ub.paragraph(
                '''
                A json query (via the jq spec) that specifies which images
                belong in the subset. Note, this is a passed as the body of
                the following jq query format string to filter valid ids
                '.images[] | select({select_images}) | .id'.

                Examples for this argument are as follows:
                '.id < 3' will select all image ids less than 3.
                '.file_name | test(".*png")' will select only images with
                file names that end with png.
                '.file_name | test(".*png") | not' will select only images
                with file names that do not end with png.
                '.myattr == "foo"' will select only image dictionaries
                where the value of myattr is "foo".
                '.id < 3 and (.file_name | test(".*png"))' will select only
                images with id less than 3 that are also pngs.
                .myattr | in({"val1": 1, "val4": 1}) will take images
                where myattr is either val1 or val4.

                Requries the "jq" python library is installed.
                '''), group='subset'),

        'select_videos': scfg.Value(
            None, help=ub.paragraph(
                '''
                A json query (via the jq spec) that specifies which videos
                belong in the subset. Note, this is a passed as the body of
                the following jq query format string to filter valid ids
                '.videos[] | select({select_images}) | .id'.

                Examples for this argument are as follows:
                '.name | startswith("foo")' will select only videos
                where the name starts with foo.

                Only applicable for dataset that contain videos.

                Requries the "jq" python library is installed.
                '''), group='subset'),


        # Histogram modifiers
        'kde': scfg.Value(True, group='histplot', help='if True compute a kernel density estimate to smooth the distribution'),
        'cumulative': scfg.Value(False, group='histplot', help='If True, plot the cumulative counts as bins increase.'),

        # 'bins': scfg.Value(256, help='Generic bin parameter that can be the name of a reference rule or the number of bins.'),
        'bins': scfg.Value('auto', group='histplot', help='Generic bin parameter that can be the name of a reference rule or the number of bins.'),

        'fill': scfg.Value(True, isflag=True, group='histplot', help='If True, fill in the space under the histogram.'),

        'element': scfg.Value('step', help='Visual representation of the histogram statistic.', group='histplot', choices=['bars', 'step', 'poly']),

        'multiple': scfg.Value('layer',
                               choices=['layer', 'dodge', 'stack', 'fill'],
                               group='histplot',
                               help='Approach to resolving multiple elements when semantic mapping creates subsets.'),

        'stat': scfg.Value('probability', choices={'count', 'frequency', 'density', 'probability'}, group='histplot', help=ub.paragraph(
            '''
            Aggregate statistic to compute in each bin.

            - ``count`` shows the number of observations
            - ``frequency`` shows the number of observations divided by the bin width
            - ``density`` normalizes counts so that the area of the histogram is 1
            - ``probability`` normalizes counts so that the sum of the bar heights is 1
            ''')),
    }


class HistAccum:
    """
    Helper to accumulate histograms
    """

    def __init__(self):
        self.accum = {}
        self.n = 0

    def update(self, data, sensor, channel):
        self.n += 1
        if sensor not in self.accum:
            self.accum[sensor] = {}
        # TODO: video / month
        sensor_accum = self.accum[sensor]
        if channel not in sensor_accum:
            sensor_accum[channel] = ub.ddict(lambda: 0)
        final_accum = sensor_accum[channel]
        for k, v in data.items():
            final_accum[k] += v

    def finalize(self):
        import pandas as pd
        import numpy as np
        # Stack all accuulated histograms into a longform dataframe
        to_stack = {}
        for sensor, sub in self.accum.items():
            for channel, hist in sub.items():
                hist = ub.sorted_keys(hist)
                # hist.pop(0)
                df = pd.DataFrame({
                    # 'intensity_bin': np.array(list(hist.keys()), dtype=int),
                    'intensity_bin': np.array(list(hist.keys())),
                    'value': np.array(list(hist.values())),
                    'channel': [channel] * len(hist),
                    'sensor': [sensor] * len(hist),
                })
                to_stack[(channel, sensor)] = df

        full_df = pd.concat(list(to_stack.values()))
        is_finite = np.isfinite(full_df.intensity_bin)
        num_nonfinite = (~is_finite).sum()
        if num_nonfinite > 0:
            import warnings
            warnings.warn('There were {} non-finite values'.format(num_nonfinite))
            full_df = full_df[is_finite]
        full_df = full_df.reset_index()
        return full_df


def main(cmdline=True, **kwargs):
    r"""
    CommandLine:
        XDEV_PROFILE=1 xdoctest -m geowatch.cli.coco_spectra main

    Example:
        >>> from geowatch.cli.coco_spectra import *  # NOQA
        >>> import kwcoco
        >>> test_dpath = ub.Path.appdir('geowatch/tests/cli/spectra').ensuredir()
        >>> image_fpath = test_dpath + '/intensityhist_demo.jpg'
        >>> coco_dset = kwcoco.CocoDataset.demo('vidshapes-msi-multisensor-videos1-frames2-gsize8')
        >>> kwargs = {'src': coco_dset, 'dst': image_fpath, 'mode': 'thread'}
        >>> kwargs['multiple'] = 'layer'
        >>> kwargs['element'] = 'step'
        >>> kwargs['workers'] = 'avail'
        >>> kwargs['show'] = False
        >>> kwargs['draw'] = False
        >>> main(cmdline=False, **kwargs)

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> from geowatch.cli.coco_spectra import *  # NOQA
        >>> import kwcoco
        >>> import geowatch
        >>> test_dpath = ub.Path.appdir('geowatch/tests/cli/spectra').ensuredir()
        >>> image_fpath = test_dpath + '/intensityhist_demo2.jpg'
        >>> coco_dset = geowatch.coerce_kwcoco('geowatch-msi')
        >>> kwargs = {
        >>>     'src': coco_dset,
        >>>     'dst': image_fpath,
        >>>     'mode': 'thread',
        >>>     'valid_range': '10:2000',
        >>>     'workers': 'avail',
        >>> }
        >>> kwargs['multiple'] = 'layer'
        >>> kwargs['element'] = 'step'
        >>> main(cmdline=False, **kwargs)
    """
    from geowatch.utils import kwcoco_extensions
    from kwutil import util_parallel
    import geowatch
    import kwcoco
    import numpy as np

    config = CocoSpectraConfig.cli(data=kwargs, cmdline=cmdline)
    print('config = {}'.format(ub.urepr(config.to_dict(), nl=1)))

    # coco_dset = kwcoco.CocoDataset.coerce(config['src'])
    coco_dset = geowatch.coerce_kwcoco(config['src'])

    valid_gids = kwcoco_extensions.filter_image_ids(
        coco_dset,
        include_sensors=config['include_sensors'],
        exclude_sensors=config['exclude_sensors'],
        select_images=config['select_images'],
        select_videos=config['select_videos'],
    )

    images = coco_dset.images(valid_gids)
    workers = util_parallel.coerce_num_workers(config['workers'])
    print('workers = {!r}'.format(workers))

    if config['max_images'] is not None:
        print('images = {!r}'.format(images))
        images = coco_dset.images(list(images)[:int(config['max_images'])])
        print('filter images = {!r}'.format(images))

    include_channels = config['include_channels']
    exclude_channels = config['exclude_channels']
    include_channels = None if include_channels is None else kwcoco.FusedChannelSpec.coerce(include_channels)
    exclude_channels = None if exclude_channels is None else kwcoco.FusedChannelSpec.coerce(exclude_channels)

    if config['valid_range'] is not None:
        valid_min, valid_max = map(float, config['valid_range'].split(':'))
    else:
        valid_min = -math.inf
        valid_max = math.inf

    class PeriodicCondition:
        """
        Helper that does something based on some periodic condition (e.g.
        number of seconds passed, or every n-th iteration)

        TODO:
            rectify with PeriodicMemoryMonitor in fusion.predict and port
            to kwutil
        """
        def __init__(self, seconds=10, do_first=True):
            self.timer = ub.Timer().tic()
            self.num_checks = 0
            self.do_first = do_first
            self.seconds = seconds

        def _satisfied(self):
            return (
                (self.timer.toc() > self.seconds) or
                (self.do_first and self.num_checks == 0)
            )

        def check(self):
            if self._satisfied():
                self.timer.tic()
                self.num_checks += 1
                return True
            return False

    jobs = ub.JobPool(mode=config['mode'], max_workers=workers, transient=True)
    from kwutil import util_progress
    pman = util_progress.ProgressManager()
    with pman:
        for coco_img in pman.progiter(images.coco_images, desc='submit stats jobs'):
            coco_img.detach()
            job = jobs.submit(ensure_intensity_stats, coco_img,
                              include_channels=include_channels,
                              exclude_channels=exclude_channels,
                              cache_dpath=config.cache_dpath,
                              bundle_dpath=coco_dset.bundle_dpath,
                              valid_min=valid_min,
                              valid_max=valid_max)
            job.coco_img = coco_img

        accum = HistAccum()

        show_seen_props = False
        if show_seen_props:
            seen_props = ub.ddict(set)

        # Update the report once every 10 seconds
        report_condition = PeriodicCondition(
            seconds=3,
            do_first=True,
        )

        for job in pman.progiter(jobs.as_completed(), total=len(jobs), desc='accumulate stats'):
            intensity_stats = job.result()
            sensor = job.coco_img.get('sensor_coarse', job.coco_img.get('sensor', 'unknown_sensor'))
            for band_stats in intensity_stats['bands']:
                intensity_hist = band_stats['intensity_hist']
                band_props = band_stats['properties']
                band_name = band_props['band_name']
                if show_seen_props:
                    for k, v in band_props.items():
                        seen_props[k].add(v)
                accum.update(intensity_hist, sensor, band_name)

            # Need to truncate repr...
            # should probably try to use geowatch.utils.util_kwutil.distributed_subitems
            # to reduce urepr overhead
            if show_seen_props:
                from kwutil.slugify_ext import smart_truncate
                seen_props_text = ub.urepr(seen_props, nl=1)
                seen_props_text = smart_truncate(seen_props_text, max_length=1600, head='\n~TRUNCATED...', tail='\n...~')
                pman.update_info(
                    ('seen_props = {}'.format(seen_props_text))
                )
            if report_condition.check():
                # TODO: can we compute a running average for efficiency
                # instead? As a workaround, only compute once every few seconds
                current_ave = single_persensor_table(accum.finalize())
                pman.update_info(
                    current_ave.to_string()
                    # ('seen_props = {}'.format(seen_props_text))
                )

        full_df = accum.finalize()
        sensor_chan_stats, distance_metrics = sensor_stats_tables(full_df)
        pman.update_info(
            sensor_chan_stats.to_string()
            # ('seen_props = {}'.format(seen_props_text))
        )

        COMPARSE_SENSORS = True
        if COMPARSE_SENSORS:
            request_columns = ['emd', 'energy_dist', 'mean_diff', 'std_diff']
            have_columns = list(ub.oset(request_columns) & ub.oset(distance_metrics.columns))
            harmony_scores = distance_metrics[have_columns].mean()
            harmony_scores = harmony_scores.to_dict()
            if harmony_scores:
                extra_text = ub.urepr(harmony_scores, precision=4, compact=1)
                print('extra_text = {!r}'.format(extra_text))
            else:
                extra_text = None
        else:
            extra_text = None

    if config['draw']:
        import kwplot
        kwplot.autosns()
        fig = plot_intensity_histograms(full_df, config)

        title_lines = []
        title = config.get('title', None)
        if title is not None:
            title_lines.append(title)

        title_lines.append(ub.Path(coco_dset.fpath).name)
        if extra_text is not None:
            title_lines.append(extra_text)

        final_title = '\n'.join(title_lines)
        fig.suptitle(final_title)

        dst_fpath = None if config['dst'] is None else ub.Path(config['dst'])
        if dst_fpath is not None:
            print('Write spectra viz to dst_fpath = {!r}'.format(dst_fpath))
            dst_fpath.parent.ensuredir()
            fig.set_size_inches(np.array([6.4, 4.8]) * 1.68)
            fig.tight_layout()
            fig.savefig(dst_fpath)

        if config['show']:
            from matplotlib import pyplot as plt
            plt.show()

    results = {
        'sensor_chan_stats': sensor_chan_stats,
        'distance_metrics': distance_metrics,
    }
    return results


def single_persensor_table(full_df):
    """
    Like sensor_stats_tables, but only used for intermediate stats

    Args:
        full_df (pd.DataFrame):
            with columns: index  intensity_bin  value channel   sensor
    """
    import pandas as pd
    import numpy as np
    sensor_channel_to_vwf = {}
    for _key, chan_df in full_df.groupby(['sensor', 'channel']):
        _sensor, channel = _key
        _values = chan_df['intensity_bin']
        _weights = chan_df['value']
        norm_weights = _weights / _weights.sum()
        sensor_channel_to_vwf[(_sensor, channel)] = {
            'raw_values': _values,
            'raw_weights': _weights,
            'norm_weights': norm_weights,
        }

    single_rows = []
    for sensor, channel in sensor_channel_to_vwf:
        sensor_data = sensor_channel_to_vwf[(sensor, channel)]
        values = sensor_data['raw_values']
        weights = sensor_data['raw_weights']

        # Note: the calculation of the variance depends on the type of
        # weighting we choose
        average = np.average(values, weights=weights)
        variance = np.average((values - average) ** 2, weights=weights)
        variance = variance * sum(weights) / (sum(weights) - 1)
        stddev = np.sqrt(variance)

        pytype = float if values.values.dtype.kind == 'f' else int

        info = {
            'min': pytype(values.min()),
            'max': pytype(values.max()),
            'mean': average,
            'std': stddev,
            'total_weight': weights.sum(),
            'channel': channel,
            'sensor': sensor,
        }
        assert info['max'] >= info['min']
        single_rows.append(info)

    sensor_chan_stats = pd.DataFrame(single_rows)
    sensor_chan_stats = sensor_chan_stats.set_index(['sensor', 'channel'])
    return sensor_chan_stats


def sensor_stats_tables(full_df):
    """
    Args:
        full_df (pd.DataFrame):
            with columns: index  intensity_bin  value channel   sensor
    """
    import itertools as it
    import scipy
    import kwarray
    import scipy.stats
    import pandas as pd
    import numpy as np
    sensor_channel_to_vwf = {}
    for _key, chan_df in full_df.groupby(['sensor', 'channel']):
        _sensor, channel = _key
        _values = chan_df['intensity_bin']
        _weights = chan_df['value']
        norm_weights = _weights / _weights.sum()
        sensor_channel_to_vwf[(_sensor, channel)] = {
            'raw_values': _values,
            'raw_weights': _weights,
            'norm_weights': norm_weights,
            'sensorchan_df': chan_df,
        }

    print('compute per-sensor stats')
    print(ub.urepr(list(sensor_channel_to_vwf.keys()), nl=1))
    single_rows = []
    for sensor, channel in ub.ProgIter(sensor_channel_to_vwf, desc='compute stats'):
        sensor_data = sensor_channel_to_vwf[(sensor, channel)]
        values = sensor_data['raw_values']
        weights = sensor_data['raw_weights']
        sensorchan_df = sensor_data['sensorchan_df']

        # Note: the calculation of the variance depends on the type of
        # weighting we choose
        average = np.average(values, weights=weights)
        variance = np.average((values - average) ** 2, weights=weights)
        variance = variance * sum(weights) / (sum(weights) - 1)
        stddev = np.sqrt(variance)

        pytype = float if values.values.dtype.kind == 'f' else int
        auto_bins = _weighted_auto_bins(
            sensorchan_df, 'intensity_bin', 'value')

        info = {
            'min': pytype(values.min()),
            'max': pytype(values.max()),
            'mean': average,
            'std': stddev,
            'total_weight': weights.sum(),
            'channel': channel,
            'sensor': sensor,
            'auto_bins': auto_bins,
        }
        assert info['max'] >= info['min']
        # print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
        single_rows.append(info)

    sensor_chan_stats = pd.DataFrame(single_rows)
    sensor_chan_stats = sensor_chan_stats.set_index(['sensor', 'channel'])
    print(sensor_chan_stats.to_string())

    print('compare channels between sensors')
    chan_to_group = ub.group_items(
        sensor_channel_to_vwf.keys(),
        [t[1] for t in sensor_channel_to_vwf.keys()]
    )
    chan_to_combos = {
        chan: list(it.combinations(group, 2)) for chan, group in chan_to_group.items()
    }
    to_compare = list(ub.flatten(chan_to_combos.values()))
    pairwise_rows = []
    for item1, item2 in ub.ProgIter(to_compare, desc='comparse_sensors', verbose=1):
        sensor1, channel1 = item1
        sensor2, channel2 = item2
        assert channel1 == channel2
        channel = channel1

        row = {
            'sensor1': sensor1,
            'sensor2': sensor2,
            'channel': channel,
        }

        u_sensor_data = sensor_channel_to_vwf[(sensor1, channel1)]
        v_sensor_data = sensor_channel_to_vwf[(sensor2, channel2)]

        u_values, u_weights = ub.take(u_sensor_data, ['raw_values', 'raw_weights'])
        v_values, v_weights = ub.take(v_sensor_data, ['raw_values', 'raw_weights'])

        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html#scipy.stats.wasserstein_distance
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.energy_distance.html#scipy.stats.energy_distance
        dist_inputs = dict(
            u_values=u_values, v_values=v_values, u_weights=u_weights,
            v_weights=v_weights)

        if 1:
            # TODO: normalize data such that density (or probability?) is 1.0
            row['emd'] = scipy.stats.wasserstein_distance(**dist_inputs)

        if 1:
            row['energy_dist'] = scipy.stats.energy_distance(**dist_inputs)

        u_stats = sensor_chan_stats.loc[(sensor1, channel1)]
        v_stats = sensor_chan_stats.loc[(sensor2, channel2)]
        row['mean_diff'] = abs(u_stats['mean'] - v_stats['mean'])
        row['std_diff'] = abs(u_stats['std'] - v_stats['std'])

        # TODO: robust alignment of pdfs
        if 1:
            cuv_values = np.union1d(u_values, v_values)
            cuv_values.sort()
            cu_weights = np.zeros(cuv_values.shape, dtype=np.float64)
            cv_weights = np.zeros(cuv_values.shape, dtype=np.float64)
            cu_indexes = np.where(kwarray.isect_flags(cuv_values, u_values))[0]
            cv_indexes = np.where(kwarray.isect_flags(cuv_values, v_values))[0]
            cu_weights[cu_indexes] = u_weights
            cv_weights[cv_indexes] = v_weights

            cu_weights * cv_weights

            row['kld'] = scipy.stats.entropy(cu_weights, cv_weights)

        pairwise_rows.append(row)

    distance_metrics = pd.DataFrame(pairwise_rows)
    print(distance_metrics.to_string())
    return sensor_chan_stats, distance_metrics


def ensure_intensity_sidecar(fpath, cache_dpath=None, bundle_dpath=None,
                             recompute=False):
    """
    Write statistics next to the image

    Args:
        fpath (str | PathLike): the path to the image to compute spectra for

        cache_dpath (str | PathLike | None):
            if specified, write cache files here instead of next to the input
            file.

        bundle_dpath (str | PathLike | None):
            if cache_dpath is specified and this is specified, this is
            considered the "root" of the image, which allows the cache to use
            specific parent folders, but not the entire filesystem.
            Specifying this allows portablity of the cache.

        recompute (bool): if True, force recomputation.

    Example:
        >>> from geowatch.cli.coco_spectra import *  # NOQA
        >>> import kwimage
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('geowatch/tests/intensity_sidecar').ensuredir()
        >>> dpath.delete().ensuredir()
        >>> img = kwimage.grab_test_image(dsize=(16, 16))
        >>> img01 = kwimage.ensure_float01(img)
        >>> img255 = kwimage.ensure_uint255(img)
        >>> fpath1 = dpath / 'img01.tif'
        >>> fpath2 = dpath / 'img255.tif'
        >>> kwimage.imwrite(fpath1, img01)
        >>> kwimage.imwrite(fpath2, img255)
        >>> fpath = fpath1
        >>> stats_fpath1 = ensure_intensity_sidecar(fpath1)
        >>> fpath = fpath2
        >>> stats_fpath2 = ensure_intensity_sidecar(fpath1)
        >>> import pickle
        >>> pickle.loads(stats_fpath1.read_bytes())
        >>> pickle.loads(stats_fpath2.read_bytes())

    Example:
        >>> from geowatch.cli.coco_spectra import *  # NOQA
        >>> import kwimage
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('geowatch/tests/intensity_sidecar2').ensuredir()
        >>> dpath.delete().ensuredir()
        >>> cache_dpath = dpath / 'cache'
        >>> bundle_dpath = None
        >>> img = kwimage.grab_test_image(dsize=(16, 16))
        >>> img255 = kwimage.ensure_uint255(img)
        >>> fpath = dpath / 'img255.tif'
        >>> kwimage.imwrite(fpath, img255)
        >>> stats_fpath = ensure_intensity_sidecar(fpath, cache_dpath=cache_dpath)
        >>> import pickle
        >>> pickle.loads(stats_fpath.read_bytes())
    """
    import os
    import kwarray
    import kwimage
    import pickle

    if cache_dpath is None:
        stats_fpath = ub.Path(os.fspath(fpath) + '.stats_v1.pkl')
    else:
        fpath = ub.Path(fpath)
        if bundle_dpath is not None:
            relpath = fpath.relative_to(bundle_dpath)
        else:
            proxy_parent = ub.hash_data(fpath.parent, hasher='sha256')
            relpath = ub.Path(proxy_parent) / fpath.name
        stats_fpath = ub.Path(cache_dpath) / (relpath + '.stats_v1.pkl')

    if recompute or not stats_fpath.exists():
        import safer
        imdata = kwimage.imread(fpath, backend='gdal', nodata_method='ma')
        imdata = kwarray.atleast_nd(imdata, 3)
        # TODO: even better float handling
        if imdata.dtype.kind == 'f':
            precision = 2
            # Round floats to p decimals to keep things semi-managable
            # binning floats is necessary due to fundamental limitations
            # We may need to decrease our resolution even further.
            imdata = imdata.round(precision)
        stats_info = {'bands': []}
        for imband in imdata.transpose(2, 0, 1):
            masked_data = imband.ravel()
            # Remove masked data
            data = masked_data.data[~masked_data.mask]
            intensity_hist = ub.dict_hist(data)
            intensity_hist = ub.sorted_keys(intensity_hist)
            stats_info['bands'].append({
                'intensity_hist': intensity_hist,
            })
        stats_fpath.parent.ensuredir()
        with safer.open(stats_fpath, 'wb') as file:
            pickle.dump(stats_info, file)
    return stats_fpath


def ensure_intensity_stats(coco_img,
                           recompute=False,
                           cache_dpath=None,
                           bundle_dpath=None,
                           include_channels=None,
                           exclude_channels=None,
                           valid_min=-math.inf,
                           valid_max=math.inf):
    """
    Ensures a sidecar file exists for the kwcoco image
    """
    from os.path import join
    import numpy as np
    import kwcoco
    import kwimage
    import pickle
    intensity_stats = {'bands': []}
    for obj in coco_img.iter_asset_objs():
        fpath = join(coco_img.bundle_dpath, obj['file_name'])
        channels = obj.get('channels', None)
        if channels is None:
            shape = kwimage.load_image_shape(fpath)
            num_channels = shape[2]
            if num_channels == 3:
                channels = 'r|g|b'
            elif num_channels == 4:
                channels = 'r|g|b|a'
            elif num_channels == 1:
                channels = 'r|g|b'
            else:
                channels = kwcoco.FusedChannelSpec.coerce(num_channels)

        channels = kwcoco.FusedChannelSpec.coerce(channels)
        declared_channel_list = channels.as_list()

        quantization = obj.get('quantization', None)

        requested_channels = channels
        if include_channels:
            requested_channels = requested_channels & include_channels
        if exclude_channels:
            requested_channels = requested_channels - exclude_channels

        if requested_channels.numel() > 0:
            stats_fpath = ensure_intensity_sidecar(fpath,
                                                   cache_dpath=cache_dpath,
                                                   bundle_dpath=bundle_dpath,
                                                   recompute=recompute)
            try:
                with open(stats_fpath, 'rb') as file:
                    stat_info = pickle.load(file)
            except Exception as ex:
                print('ex = {!r}'.format(ex))
                print('error loading stats_fpath = {!r}'.format(stats_fpath))
                raise

            alwaysappend = requested_channels.numel() == channels.numel()

            for band_idx, band_stat in enumerate(stat_info['bands']):
                try:
                    band_name = declared_channel_list[band_idx]
                except IndexError:
                    print('bad channel declaration fpath = {!r}'.format(fpath))
                    if 0:
                        print('obj = {}'.format(ub.urepr(obj, nl=1)))
                        print('coco_img = {!r}'.format(coco_img))
                        print('fpath = {!r}'.format(fpath))
                        print('stats_fpath = {!r}'.format(stats_fpath))
                        print(len(stat_info['bands']))
                        print('band_idx = {!r}'.format(band_idx))
                        print('channels = {!r}'.format(channels))
                    # raise
                    band_name = 'unknown'
                if quantization is not None:
                    # Handle dequantization
                    from delayed_image.helpers import dequantize
                    quant_values = np.array(list(band_stat['intensity_hist'].keys()))
                    counts = list(band_stat['intensity_hist'].values())
                    dequant_values = dequantize(quant_values, quantization)
                    band_stat['intensity_hist'] = ub.odict(zip(dequant_values, counts))
                if alwaysappend or band_name in requested_channels:
                    band_stat['properties'] = props = {
                        'video_name': None,
                        'band_name': band_name,
                        'month': None,
                    }
                    try:
                        props['video_name'] = coco_img.video['name']
                    except Exception:
                        ...
                    try:
                        from kwutil import util_time
                        props['month'] = util_time.coerce_datetime(coco_img.img['date_captured']).month
                    except Exception:
                        ...
                    band_stat['band_name'] = band_name
                    intensity_stats['bands'].append(band_stat)

    if math.isfinite(valid_max) or math.isfinite(valid_min):
        # Filter on the fly, but in the worker process
        for band_stats in intensity_stats['bands']:
            intensity_hist = band_stats['intensity_hist']
            band_stats['intensity_hist'] = {
                k: v for k, v in intensity_hist.items()
                if k >= valid_min and k <= valid_max}

    return intensity_stats


def plot_intensity_histograms(full_df, config, ax=None):
    """
    Args:
        full_df (pd.DataFrame):
            with columns: index  intensity_bin  value channel   sensor
        ax (Axes | None):
            if specified, we assume only 1 plot is made

    Example:
        >>> # xdoctest: +REQUIRES(--show)
        >>> from geowatch.cli.coco_spectra import *  # NOQA
        >>> import pandas as pd
        >>> n = 256
        >>> data = {
        >>>     'intensity_bin': np.arange(0, n),
        >>>     'value': np.random.randint(0, 256, n),
        >>>     'channel': ['red'] * n,
        >>>     'sensor': ['unknown'] * n,
        >>> }
        >>> full_df = pd.DataFrame(data)
        >>> config = CocoSpectraConfig()
        >>> kwplot.autompl()
        >>> ax = plot_intensity_histograms(full_df, config)
    """
    unique_channels = full_df['channel'].unique()
    unique_sensors = full_df['sensor'].unique()

    import kwplot
    sns = kwplot.autosns()

    palette = {
        'red': 'red',
        'blue': 'blue',
        'green': 'green',
        'cirrus': 'skyblue',
        'coastal': 'purple',
        'nir': 'orange',
        'swir16': 'pink',
        'swir22': 'hotpink',
    }
    if 'red' not in unique_channels and 'r' in unique_channels:
        # hack
        palette['r'] = 'red'
        palette['g'] = 'green'
        palette['b'] = 'blue'
    for channel in unique_channels:
        if channel not in palette:
            palette[channel] = None
    palette = _fill_missing_colors(palette)

    hist_data_kw = dict(
        x='intensity_bin',
        weights='value',
        bins=config['bins'],
        stat=config['stat'],
        hue='channel',
    )
    hist_style_kw = dict(
        palette=palette,
        fill=config['fill'],
        element=config['element'],
        multiple=config['multiple'],
        kde=config['kde'],
        cumulative=config['cumulative'],
    )

    if 0:
        # __hisplot_notes__
        import inspect
        sig = inspect.signature(sns.histplot)
        # Print params we might not have looked at in detail
        exposed_params = {
            'cumulative', 'kde', 'multiple', 'element', 'fill', 'hue', 'stat',
            'bins', 'weights', 'x', 'palette',
        }
        probably_ignorable_params = {
            'pmax', 'hue_order', 'hue_norm', 'cbar', 'cbar_kws', 'cbar_ax', 'ax',
            'legend', 'thresh' 'y',
        }
        maybe_expose = (set(sig.parameters) - exposed_params) - probably_ignorable_params
        print('maybe_expose = {}'.format(ub.urepr(maybe_expose, nl=1)))

    # For S2 that is supposed to be divide by 10000.
    # For L8 it is multiply by 2.75e-5 and subtract 0.2.
    # 1 / 2.75e-5
    print('start plot')

    print(f'unique_sensors={unique_sensors}')
    if ax is None:
        fig = kwplot.figure(fnum=1, doclf=True)
        fig.clf()
        print('fig = {!r}'.format(fig))
        pnum_ = kwplot.PlotNums(nSubplots=len(unique_sensors))
        print('pnum_ = {!r}'.format(pnum_))
    else:
        fig = None
        assert len(unique_sensors) == 1

    for sensor_name, sensor_df in full_df.groupby('sensor'):
        print('plot sensor_name = {!r}'.format(sensor_name))

        hist_data_kw_ = hist_data_kw.copy()
        if hist_data_kw_['bins'] == 'auto':
            xvar = hist_data_kw['x']
            weightvar = hist_data_kw['weights']
            hist_data_kw_['bins'] = _weighted_auto_bins(sensor_df, xvar, weightvar)

        ax = kwplot.figure(fnum=1, pnum=pnum_()).gca()

        # z = [tuple(a.values()) for a in sensor_df[['intensity_bin', 'channel', 'sensor']].to_dict('records')]
        # ub.find_duplicates(z)
        try:
            # We have already computed the histogram, but we can get seaborn to
            # show it using a simple trick: use weight to represent the
            # frequency that should be used for every bin.

            # https://github.com/mwaskom/seaborn/issues/2709
            sns.histplot(ax=ax, data=sensor_df.reset_index(), **hist_data_kw_, **hist_style_kw)
        except Exception:
            print('hist_data_kw_ = {}'.format(ub.urepr(hist_data_kw_, nl=1)))
            print('hist_style_kw = {}'.format(ub.urepr(hist_style_kw, nl=1)))
            print('ERROR')
            print(sensor_df)
            raise
            pass
        if config['valid_range'] is not None:
            valid_min, valid_max = map(float, config['valid_range'].split(':'))
            ax.set_xlim(valid_min, valid_max)

        ax.set_title(sensor_name)
        # maxx = sensor_df.intensity_bin.max()
        # maxx = sensor_maxes[sensor_name]
        # ax.set_xlim(0, maxx)

    return fig


def _weighted_auto_bins(data, xvar, weightvar):
    """
    Generalized histogram bandwidth estimators for weighted univariate data

    References:
        https://github.com/mwaskom/seaborn/issues/2710

    TODO:
        add to util_kwarray

    Example:
        >>> import pandas as pd
        >>> import numpy as np
        >>> n = 100
        >>> to_stack = []
        >>> rng = np.random.RandomState(432)
        >>> for group_idx in range(3):
        >>>     part_data = pd.DataFrame({
        >>>         'x': np.arange(n),
        >>>         'weights': rng.randint(0, 100, size=n),
        >>>         'hue': [f'group_{group_idx}'] * n,
        >>>     })
        >>>     to_stack.append(part_data)
        >>> data = pd.concat(to_stack).reset_index()
        >>> xvar = 'x'
        >>> weightvar = 'weights'
        >>> n_equal_bins = _weighted_auto_bins(data, xvar, weightvar)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import seaborn as sns
        >>> sns.histplot(data=data, bins=n_equal_bins, x='x', weights='weights', hue='hue')
    """
    import numpy as np
    sort_df = data.sort_values(xvar)
    values = sort_df[xvar]
    weights = sort_df[weightvar]
    minval = values.iloc[0]
    maxval = values.iloc[-1]

    total = weights.sum()
    ptp = maxval - minval

    # _hist_bin_sqrt = ptp / np.sqrt(total)
    _hist_bin_sturges = ptp / (np.log2(total) + 1.0)

    cumtotal = weights.cumsum().values
    quantiles = cumtotal / cumtotal[-1]
    idx2, idx1 = np.searchsorted(quantiles, [0.75, 0.25])
    # idx2, idx1 = _weighted_quantile(weights, [0.75, 0.25])
    iqr = values.iloc[idx2] - values.iloc[idx1]
    _hist_bin_fd = 2.0 * iqr * total ** (-1.0 / 3.0)

    fd_bw = _hist_bin_fd  # Freedman-Diaconis
    sturges_bw = _hist_bin_sturges

    if fd_bw:
        bw_est = min(fd_bw, sturges_bw)
    else:
        # limited variance, so we return a len dependent bw estimator
        bw_est = sturges_bw

    # from numpy.lib.histograms import _get_outer_edges, _unsigned_subtract
    first_edge, last_edge = _get_outer_edges(values, None)
    if bw_est:
        n_equal_bins = int(np.ceil(_unsigned_subtract(last_edge, first_edge) / bw_est))
    else:
        # Width can be zero for some estimators, e.g. FD when
        # the IQR of the data is zero.
        n_equal_bins = 1

    # Take the minimum of this and the number of actual bins
    n_equal_bins = min(n_equal_bins, len(values))
    return n_equal_bins


def _get_outer_edges(a, range):
    """
    Determine the outer bin edges to use, from either the data or the range
    argument

    Note: vendored from numpy.lib._histograms_impl
    """
    import numpy as np
    if range is not None:
        first_edge, last_edge = range
        if first_edge > last_edge:
            raise ValueError(
                'max must be larger than min in range parameter.')
        if not (np.isfinite(first_edge) and np.isfinite(last_edge)):
            raise ValueError(
                "supplied range of [{}, {}] is not finite".format(first_edge, last_edge))
    elif a.size == 0:
        # handle empty arrays. Can't determine range, so use 0-1.
        first_edge, last_edge = 0, 1
    else:
        first_edge, last_edge = a.min(), a.max()
        if not (np.isfinite(first_edge) and np.isfinite(last_edge)):
            raise ValueError(
                "autodetected range of [{}, {}] is not finite".format(first_edge, last_edge))

    # expand empty range to avoid divide by zero
    if first_edge == last_edge:
        first_edge = first_edge - 0.5
        last_edge = last_edge + 0.5

    return first_edge, last_edge


def _unsigned_subtract(a, b):
    """
    Subtract two values where a >= b, and produce an unsigned result

    This is needed when finding the difference between the upper and lower
    bound of an int16 histogram

    Note: vendored from numpy.lib._histograms_impl
    """
    import numpy as np
    # coerce to a single type
    signed_to_unsigned = {
        np.byte: np.ubyte,
        np.short: np.ushort,
        np.intc: np.uintc,
        np.int_: np.uint,
        np.longlong: np.ulonglong
    }
    dt = np.result_type(a, b)
    try:
        unsigned_dt = signed_to_unsigned[dt.type]
    except KeyError:
        return np.subtract(a, b, dtype=dt)
    else:
        # we know the inputs are integers, and we are deliberately casting
        # signed to unsigned.  The input may be negative python integers so
        # ensure we pass in arrays with the initial dtype (related to NEP 50).
        return np.subtract(np.asarray(a, dtype=dt), np.asarray(b, dtype=dt),
                           casting='unsafe', dtype=unsigned_dt)


def _fill_missing_colors(label_to_color):
    """
    label_to_color = {'foo': kwimage.Color('red').as01(), 'bar': None}
    """
    from distinctipy import distinctipy
    import kwarray
    import numpy as np
    import kwimage
    given = {k: kwimage.Color(v).as01() for k, v in label_to_color.items() if v is not None}
    needs_color = sorted(set(label_to_color) - set(given))

    seed = 6777939437
    # hack in our code

    def _patched_get_random_color(pastel_factor=0, rng=None):
        rng = kwarray.ensure_rng(seed, api='python')
        color = [(rng.random() + pastel_factor) / (1.0 + pastel_factor) for _ in range(3)]
        return tuple(color)
    distinctipy.get_random_color = _patched_get_random_color

    exclude_colors = [
        tuple(map(float, (d, d, d)))
        for d in np.linspace(0, 1, 5)
    ] + list(given.values())

    final = given.copy()
    new_colors = distinctipy.get_colors(len(needs_color), exclude_colors=exclude_colors)
    for key, new_color in zip(needs_color, new_colors):
        final[key] = tuple(map(float, new_color))
    return final


__cli__ = CocoSpectraConfig

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os
from geowatch.cli.smartflow_ingress import smartflow_ingress
from geowatch.cli.smartflow_egress import smartflow_egress
from geowatch.utils.util_framework import download_region
import ubelt as ub
import scriptconfig as scfg
from geowatch.mlops import smart_pipeline
from kwutil.util_yaml import Yaml


_debug_notes_ = r"""

docker run \
    --runtime=nvidia \
    --volume "$HOME/temp/debug_smartflow_v2/ingress":/tmp/ingress \
    --volume $HOME/.aws:/root/.aws:ro \
    --volume "$HOME/code":/extern_code:ro \
    --volume "$HOME/data":/extern_data:ro \
    --volume "$HOME"/.cache/pip:/pip_cache \
    --env AWS_PROFILE=iarpa \
    -it registry.smartgitlab.com/kitware/watch:0.12.1-98697830f-strict-pyenv3.11.2-20231112T181927-0500-from-de082898 bash

ipython

from geowatch.cli.smartflow.run_sv_datagen import *  # NOQA
config = {
    'input_path'        : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v134/batch/kit/KR_R001/2021-08-31/split/mono/products/bas-fusion/items.jsonl',
    'input_region_path' : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v134/batch/kit/KR_R001/2021-08-31/input/mono/region_models/KR_R001.geojson',
    'output_path'       : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v134/batch/kit/KR_R001/2021-08-31/split/mono/products/site-cropped-kwcoco-for-sv/items.jsonl',
    'aws_profile'       : None,
    'dryrun'            : False,
    'outbucket'         : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v134/batch/kit/KR_R001/2021-08-31/split/mono/products/site-cropped-kwcoco-for-sv',
    'newline'           : True,
    'dont_recompute'    : False,
    'sv_cropping_config': 'context_factor: 1.6\nforce_min_gsd: 1GSD\nminimum_size: 256x256@3GSD\nnum_end_frames: 3.0\nnum_start_frames: 3.0\ntarget_gsd: 2GSD',
}
config = SVDatasetConfig(**config)

"""


class SVDatasetConfig(scfg.DataConfig):
    """
    Generate cropped KWCOCO dataset for SC
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
        '''
        Output as simple newline separated STAC items
        '''))
    dont_recompute = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Will not recompute if output_path already exists
        '''))
    sv_cropping_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for SV_Cropping.
        '''))

    input_region_models_asset_name = scfg.Value('cropped_region_models_bas', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to use as input
        '''))

    input_site_models_asset_name = scfg.Value('cropped_site_models_bas', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to to use as input
        '''))


def main():
    config = SVDatasetConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_generate_sv_cropped_kwcoco(config)


def run_generate_sv_cropped_kwcoco(config):
    from geowatch.utils import util_framework
    from geowatch.utils import util_fsspec

    input_path = config.input_path
    input_region_path = config.input_region_path
    output_path = config.output_path
    outbucket = config.outbucket
    aws_profile = config.aws_profile
    dryrun = config.dryrun

    # newline = config.newline

    dont_recompute = config.dont_recompute
    sv_cropping_config = config.sv_cropping_config

    if aws_profile is not None:
        # This should be sufficient, but it is not tested.
        util_fsspec.S3Path._new_fs(profile=aws_profile)

    ####
    # DEBUGGING:
    # Print info about what version of the code we are running on
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    if dont_recompute:
        output_path = util_fsspec.FSPath.coerce(output_path)
        if output_path.exists():
            # If output_path file was there, nothing to do
            return

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')
    ingressed_assets = smartflow_ingress(
        input_path=input_path,
        assets=[
            'kwcoco_for_sc',
            config.input_region_models_asset_name,
            config.input_site_models_asset_name,
        ],
        outdir=ingress_dir,
        aws_profile=aws_profile,
        dryrun=dryrun)

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    local_region_path = download_region(input_region_path,
                                        local_region_path,
                                        aws_profile=aws_profile,
                                        strip_nonregions=True)

    # Parse region_id from original region file
    region_id = util_framework.determine_region_id(local_region_path)

    if region_id is None:
        raise RuntimeError("Couldn't parse 'region_id' from input region file")

    # Paths to inputs generated in previous pipeline steps
    bas_region_path = ub.Path(ingressed_assets[config.input_region_models_asset_name]) / f'{region_id}.geojson'
    ta1_sc_kwcoco_path = ingressed_assets['kwcoco_for_sc']

    node_state.print_current_state(ingress_dir)

    # 4. Crop ingress KWCOCO dataset to region for SV
    print("* Cropping KWCOCO dataset to region for SV*")
    ta1_sv_cropped_kwcoco_path = ingress_dir / 'cropped_kwcoco_for_sv.json'

    sv_cropping_config = Yaml.coerce(sv_cropping_config or {})

    sv_cropping = smart_pipeline.SV_Cropping(root_dpath=ingress_dir)
    sv_cropping.configure(sv_cropping_config | {
        'crop_src_fpath': ta1_sc_kwcoco_path,
        'regions': bas_region_path,
        'crop_dst_fpath': ta1_sv_cropped_kwcoco_path})

    sv_crop_command = sv_cropping.command()
    print(sv_crop_command)
    ub.cmd(sv_crop_command, check=True, verbose=3, system=True)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)

    assets_dpath = ingress_dir / f'{region_id}'
    assets_dpath.ensuredir()

    # Ensure at least one file exists in the directory we want to upload
    # so aws doesn't break.
    # TODO: we have standardized on "__dir__" being the dummy filename.
    # verify that this doesn't break anything and change it here.
    (assets_dpath / '_empty').touch()

    print("* Egressing KWCOCO dataset and associated STAC item *")
    ingressed_assets['cropped_kwcoco_for_sv'] = ta1_sv_cropped_kwcoco_path
    ingressed_assets['cropped_kwcoco_for_sv_assets'] = ingress_dir / f'{region_id}'
    # Ensure that the assets directory at least exists before copying
    # (for incremental mode it's possible there are no assets for the
    # given interval; egress will fail if a source directory / file
    # doesn't exist)
    if not os.path.isdir(ingressed_assets['cropped_kwcoco_for_sv_assets']):
        os.makedirs(ingressed_assets['cropped_kwcoco_for_sv_assets'], exist_ok=True)
        ub.Path(ingressed_assets['cropped_kwcoco_for_sv_assets'] / '.empty').touch()

    node_state.print_current_state(ingress_dir)

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     output_path,
                     outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False)


if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements convolutional and attentional residual block groups."""

import math

import tensorflow as tf
from deeplab2.model import utils
from deeplab2.model.layers import activations
from deeplab2.model.layers import axial_blocks
from deeplab2.model.layers import drop_path
from deeplab2.model.layers import dual_path_transformer
from deeplab2.model.layers import positional_encodings
from deeplab2.model.layers import recompute_grad as recompute_grad_lib

# We will apply 10x larger learning rates on transformer layers. This global
# variable name will be accessed when we build the optimizers. This keyword is
# reserved and should not be a part of the variable names in a classification
# pretrained backbone.
TRANSFORMER = 'transformer'


def _get_current_names(index):
    current_name = '_block{}'.format(index + 1)
    transformer_current_name = '_block{}_{}'.format(index + 1, TRANSFORMER)
    return current_name, transformer_current_name


class BlockGroup(tf.keras.layers.Layer):
    """Applies a group of residual blocks with dual path transformer layers [1].

    An optional dual-path transformer layer is inserted after each residual block.
    The transformer layer performs memory2pixel attention, pixel2memory attention,
    and memory2memory self-attention, while the standard residual block applies
    the pixel2pixel axial-attention, global-attention, or spatial convolution.

    Reference:
    [1] MaX-DeepLab: End-to-End Panoptic Segmentation with Mask Transformers,
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    def __init__(self,
                 filters,
                 num_blocks,
                 name,
                 original_resnet_stride,
                 original_resnet_input_stride,
                 output_stride=16,
                 backbone_type='resnet_beta',
                 positional_encoding_type=None,
                 use_global_beyond_stride=0,
                 use_axial_beyond_stride=16,
                 use_transformer_beyond_stride=32,
                 use_sac_beyond_stride=0,
                 use_squeeze_and_excite=False,
                 conv_use_recompute_grad=False,
                 axial_use_recompute_grad=True,
                 recompute_within_stride=0,
                 transformer_use_recompute_grad=False,
                 transformer_expansion=1,
                 drop_path_keep_prob=0.8,
                 drop_path_beyond_stride=16,
                 drop_path_schedule='constant',
                 activation='relu',
                 attention_bottleneck_expansion=2,
                 axial_layer_config=None,
                 dual_path_transformer_layer_config=None,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay=0.0,
                 auxiliary_predictor_func=None,
                 use_axial_block=True):
        """Initializes a BlockGroup layer.

        Args:
          filters: An integer, the base number of channels for this block group.
          num_blocks: An integer, the number of blocks for this block group.
          name: A string, the name of the block group.
          original_resnet_stride: An integer, the original resnet stride for this
            block, usually 1 or 2. The stride will be applied if
            original_resnet_input_stride is smaller than the desired output_stride.
            Otherwise, the stride will not be applied, and atrous convolution will
            be used after the first block.
          original_resnet_input_stride: An integer, the total input stride in the
            original resnet. For example, the total input stride for the last stage
            of the original resnet is 16, and the total output stride is 32. This
            stride differs from the true stride of the feature in that we might use
            atrous convolution to change both the input and output stride to, e.g.
            8, but its original resnet input stride remains the same. In this case,
            we also use the original resnet input stride to compute the atrous rate.
          output_stride: An integer, the desired output_stride for the ResNet.
          backbone_type: A string, the type of the backbone. Supports 'resnet',
            'resnet_beta', and 'wider_resnet'. The 'resnet' refers to the original
            resnet with a 7x7 convolutional stem. The 'resnet_beta' means a resnet
            but with an inception stem. The 'wider_resnet' is a wider variant of
            resnet with extensively used 3x3 convolutions.
          positional_encoding_type: A string, type of the positional encoding.
            Support '2D', '1D', and None.
          use_global_beyond_stride: An integer, the stride beyond which we use
            global attention. Set to 0 if no global attention is desired. Defaults
            to 0, i.e. we do not use global attention.
          use_axial_beyond_stride: An integer, the stride beyond which we use axial
            attention. Note that use_global_beyond_stride has a higher priority,
            i.e. we use global attention if the stride is also beyond
            use_global_beyond_stride. Set to 0 if no axial attention is desired.
            Defaults to 16 as in MaX-DeepLab.
          use_transformer_beyond_stride: An integer, the stride beyond which we use
            a transformer layer. Set to 0 if no transformer is desired. Defaults to
            32 as in MaX-DeepLab-S.
          use_sac_beyond_stride: An integer. Use the Switchable Atrous Convolution
            (SAC) beyond the specified stride. For example, if
            `use_sac_beyond_stride` = 16, SAC will be applied to the network stage
            whose output stride >= 16 (i.e., 16 and 32). Set to 0 or -1 to disable
            it. Defaults to 0 as SAC is not used in MaX-DeepLab.
          use_squeeze_and_excite: A boolean, whether squeeze-and-excite (SE) is
            used. Defaults to False as SE is not used in MaX-DeepLab.
          conv_use_recompute_grad: A boolean, whether to use the gradient
            checkpointing trick for convolutional blocks. This trick reduces
            accelerator memory usage, but takes longer to compute gradients.
            Defaults to False since convolutional layers are memory efficient.
          axial_use_recompute_grad: A boolean, whether to use the gradient
            checkpointing trick for axial blocks. This trick reduces accelerator
            memory usage, but takes longer to compute gradients. Defaults to True
            since it saves memory for axial blocks.
          recompute_within_stride: An integer, the stride within which we use the
            gradient checkpointing trick. This trick reduces accelerator memory
            usage, but takes longer to compute gradients. Defaults to 0 (do not
            recompute any layer).
          transformer_use_recompute_grad: A boolean, whether to use the gradient
            checkpointing trick for dual-path transformer blocks. This trick reduces
            accelerator memory usage, but takes longer to compute gradients.
            Defaults to False.
          transformer_expansion: An integer, the expansion ratio for the transformer
            bottleneck.
          drop_path_keep_prob: A float, the keep probability for dropping path.
            Defaults to 0.8 as in MaX-DeepLab-S.
          drop_path_beyond_stride: An integer, the stride beyond which we apply drop
            path augmentation. Defaults to 16 as in MaX-DeepLab-S.
          drop_path_schedule: A string, the drop path schedule. Currently, we
            support 'constant': use the same drop path keep probability for all
            stages, and 'linear': linearly decrease the drop path keep probability
            from 1.0 at 0-th stage (or STEM) to `drop_path_keep_prob` at last stage.
          activation: A string, type of activation function to apply. Support
            'relu', 'swish' (or 'silu'), 'gelu', 'approximated_gelu', and 'elu'.
          attention_bottleneck_expansion: An integer, the expansion ratio for
            axial attention blocks.
          axial_layer_config: A dict, an argument dictionary for the axial layer.
          dual_path_transformer_layer_config: A dict, an argument dictionary for the
            transformer.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.
          auxiliary_predictor_func: A callable function that returns an
            initialization of auxiliary predictor.
          use_axial_block: A boolean. When set to False, the axial block will be
            disabled. This is helpful when we just want to perform
            dual-path transformer (i.e., no pixel2pixel attention). Default to True.

        Raises:
          ValueError: If backbone_type is not one of 'resnet', 'resnet_beta', or
            'wider_resnet'.
          ValueError: original_resnet_input_stride is not power of 2.
          ValueError: output_stride is not power of 2.
        """
        if original_resnet_input_stride & (original_resnet_input_stride - 1):
            raise ValueError('original_resnet_input_stride is not power of 2.')
        if output_stride & (output_stride - 1):
            raise ValueError('output_stride is not power of 2.')

        super(BlockGroup, self).__init__(name=name)
        self._add_absolute_positional_encoding = None
        self._activation_fn = activations.get_activation(activation)
        self._num_blocks = num_blocks
        self._drop_path_keep_prob = []
        self._recompute_grad = []
        self._transformer_use_recompute_grad = transformer_use_recompute_grad
        self._use_axial_block = use_axial_block
        if dual_path_transformer_layer_config is None:
            dual_path_transformer_layer_config = {}
        original_resnet_current_stride = original_resnet_input_stride

        use_sac = (original_resnet_input_stride * original_resnet_stride >=
                   use_sac_beyond_stride > 0)

        recompute_grad = (original_resnet_input_stride * original_resnet_stride <=
                          recompute_within_stride)

        for index in range(num_blocks):
            current_name, transformer_current_name = _get_current_names(index)

            # Compute the current strides. If there is a stride for this block group,
            # we do it in the first residual block.
            if index == 0 and original_resnet_input_stride < output_stride:
                current_strides = original_resnet_stride
            else:
                current_strides = 1

            # Compute the current atrous rate.
            if original_resnet_current_stride > output_stride:
                atrous_rate = original_resnet_current_stride // output_stride
            else:
                atrous_rate = 1

            # Compute the atrous rate for the second conv in the first basic block.
            if (index == 0 and original_resnet_input_stride * original_resnet_stride >
                    output_stride):
                basic_block_second_conv_atrous_rate = (
                        original_resnet_input_stride * original_resnet_stride //
                        output_stride)
            else:
                basic_block_second_conv_atrous_rate = atrous_rate

            # Compute the current drop_path_keep_prob.
            current_stage = math.log2(original_resnet_current_stride) - 1
            if original_resnet_current_stride >= drop_path_beyond_stride:
                current_drop_path_keep_prob = drop_path.get_drop_path_keep_prob(
                    drop_path_keep_prob, drop_path_schedule,
                    current_stage=int(round(current_stage)),
                    num_stages=4)
            else:
                current_drop_path_keep_prob = 1.0

            # Compute which block_fn to use for this residual block.
            if original_resnet_current_stride >= use_global_beyond_stride > 0:
                attention_type = 'global'
                recompute_grad = axial_use_recompute_grad or recompute_grad
                filters_list = [filters * attention_bottleneck_expansion,
                                filters,
                                filters * 4]
            elif original_resnet_current_stride >= use_axial_beyond_stride > 0:
                attention_type = 'axial'
                recompute_grad = axial_use_recompute_grad or recompute_grad
                filters_list = [filters * attention_bottleneck_expansion,
                                filters,
                                filters * 4]
            elif backbone_type == 'resnet' or backbone_type == 'resnet_beta':
                attention_type = None
                recompute_grad = conv_use_recompute_grad or recompute_grad
                filters_list = [filters,
                                filters,
                                filters * 4]
            elif backbone_type == 'wider_resnet':
                if original_resnet_input_stride * original_resnet_stride < 32:
                    # Wider-ResNet uses conv basic blocks except the last stage.
                    attention_type = None
                    recompute_grad = conv_use_recompute_grad or recompute_grad
                    filters_list = [filters * 4,
                                    filters * 4]
                else:
                    # Wider-ResNet uses an expanded bottleneck block in the last stage.
                    attention_type = None
                    recompute_grad = conv_use_recompute_grad or recompute_grad
                    filters_list = [filters,
                                    filters * 2,
                                    filters * 4]
            else:
                raise ValueError(backbone_type + ' is not supported.')

            self._drop_path_keep_prob.append(current_drop_path_keep_prob)
            # Apply the residual block.
            block_fn = None
            if use_axial_block:
                # The inputs to block_fn should be activated features.
                block_fn = axial_blocks.AxialBlock(
                    filters_list,
                    kernel_size=3,
                    strides=current_strides,
                    atrous_rate=atrous_rate,
                    use_squeeze_and_excite=use_squeeze_and_excite,
                    use_sac=use_sac,
                    bn_layer=bn_layer,
                    activation=activation,
                    name=current_name[1:],
                    conv_kernel_weight_decay=conv_kernel_weight_decay,
                    basic_block_second_conv_atrous_rate=(
                        basic_block_second_conv_atrous_rate),
                    attention_type=attention_type,
                    axial_layer_config=axial_layer_config)
            self._recompute_grad.append(recompute_grad)
            utils.safe_setattr(self, current_name, block_fn)

            # Modify the original_resnet_stride according to the strides.
            if index == 0 and original_resnet_stride > 1:
                original_resnet_current_stride *= original_resnet_stride
                # Add absolute positional encoding if we will apply global attention
                # beyond this stride.
                if original_resnet_current_stride == use_global_beyond_stride > 0:
                    self._add_absolute_positional_encoding = (
                        positional_encodings.AddAbsolutePositionalEncoding(
                            'add_absolute_positional_encoding',
                            positional_encoding_type, bn_layer, conv_kernel_weight_decay))
            if original_resnet_current_stride >= use_transformer_beyond_stride > 0:
                # Apply a dual-path transformer.
                transformer_block_fn = dual_path_transformer.DualPathTransformerLayer(
                    name=transformer_current_name[1:],
                    filters=int(128 * transformer_expansion),
                    activation=activation,
                    bn_layer=bn_layer,
                    conv_kernel_weight_decay=conv_kernel_weight_decay,
                    auxiliary_predictor_func=auxiliary_predictor_func,
                    **dual_path_transformer_layer_config)
                utils.safe_setattr(self, transformer_current_name, transformer_block_fn)
            else:
                utils.safe_setattr(self, transformer_current_name, None)
        # Avoid using recompute_grad for the first call that builds the sub-layers.
        # Otherwise, recompute_grad will not track newly built model parameters.
        self._first_building_call = True

    def call(self, inputs, training=False):
        """Performs a forward pass.

        Args:
          inputs: A list of tensors or tuples. The first tensor is a
            pixel_space_input with shape [batch, height, width, pixel_channels].
            The second tensor is memory_space_input with shape [batch, length,
            memory_channels]. The third one is an optional auxiliary_outputs which
            is a tuple containing auxiliary outputs, where each element has the
            dictionary type.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          output: An output [batch, height, width, filters * 4] tensor.
          activated_output: An activated output [batch, height, width, filters * 4]
            tensor.
          memory_space_output: A memory space output [batch, length,
            memory_channels] tensor.
          auxiliary_outputs: An optional tuple containing auxiliary outputs, where
            each element has the dictionary type.

        Raises:
          ValueError: If the length of inputs is not 2 or 3.
        """
        # The pixel space inputs are non-activated features.
        if len(inputs) == 2:
            features, memory_space_output = inputs
            auxiliary_outputs = ()
            return_auxiliary_outputs = False
        elif len(inputs) == 3:
            features, memory_space_output, auxiliary_outputs = inputs
            return_auxiliary_outputs = True
        else:
            raise ValueError('The length of inputs should be either 2 or 3!')

        # Recompute_grad takes only float tensors as inputs. It does not allow
        # bools or boolean tensors. For this reason, we cast training to a float
        # tensor and cast it back after we go through the recompute_grad wrap.
        float_tensor_training = tf.cast(training, tf.float32)

        for index in range(self._num_blocks):
            current_name, transformer_current_name = _get_current_names(index)
            block_fn_no_recompute = getattr(
                self, current_name)
            transformer_block_fn_no_recompute = getattr(
                self, transformer_current_name)
            current_drop_path_keep_prob = self._drop_path_keep_prob[index]

            if self._use_axial_block:
                # Wrap the layer if we want to recompute it in the backward pass.
                if (self._recompute_grad[index] and training):
                    # The seed is not actually used since we do not have any random
                    # operation in the recomputed function. The purpose of the provided
                    # seed is to prevent recompute_grad from generating a new seed
                    # variable which is not compatible with model exporting.
                    block_fn = recompute_grad_lib.recompute_grad(
                        block_fn_no_recompute, seed=tf.constant(0, tf.int32))
                else:
                    block_fn = block_fn_no_recompute

                # The inputs to block_fn should be activated features.
                block_fn_inputs = [features, float_tensor_training]
                # We have to define drop_path_masks outside the layer call and pass it
                # into the layer, because tf.recompute_grad (gradient checkpointing)
                # does not allow any randomness within the function call. In addition,
                # recompute_grad functions can only take Tensors as inputs, so we do not
                # pass the drop_path_random_mask (when it is None) into block_fn.
                if current_drop_path_keep_prob < 1.0 and training:
                    drop_path_random_mask = drop_path.generate_drop_path_random_mask(
                        features, current_drop_path_keep_prob)

                    block_fn_inputs.append(drop_path_random_mask)

                # Build the sub-layers when the block_fn is called for the first time.
                # Otherwise, recompute_grad will not track newly built model parameters.
                if self._first_building_call:
                    _ = block_fn_no_recompute(tuple(block_fn_inputs))
                # Apply the residual block.
                features = block_fn(tuple(block_fn_inputs))

            if index == 0 and self._add_absolute_positional_encoding is not None:
                features = self._add_absolute_positional_encoding(features,
                                                                  training=training)

            if transformer_block_fn_no_recompute is not None:
                # Wrap the layer if we want to recompute it in the backward pass.
                if (self._transformer_use_recompute_grad and training):
                    # The seed is not actually used since we do not have any random
                    # operation in the recomputed function. The purpose of the provided
                    # seed is to prevent recompute_grad from generating a new seed
                    # variable which is not compatible with model exporting.
                    transformer_block_fn = recompute_grad_lib.recompute_grad(
                        transformer_block_fn_no_recompute, seed=tf.constant(0, tf.int32))
                else:
                    transformer_block_fn = transformer_block_fn_no_recompute

                transformer_block_fn_input_list = [
                    features, memory_space_output, auxiliary_outputs,
                    float_tensor_training]
                # We have to define drop_path_masks outside the layer call and pass it
                # into the layer, because recompute_grad (gradient checkpointing) does
                # not allow any randomness within the function call. In addition,
                # recompute_grad functions can only take Tensors as inputs, so we do not
                # pass the drop_path_masks (when they are None) into
                # transformer_block_fn.
                if current_drop_path_keep_prob < 1.0 and training:
                    # Drop path random mask for pixel space attention.
                    pixel_space_drop_path_mask = drop_path.generate_drop_path_random_mask(
                        memory_space_output, current_drop_path_keep_prob)
                    # Drop path random mask for memory space attention.
                    memory_space_attention_drop_path_mask = (
                        drop_path.generate_drop_path_random_mask(
                            memory_space_output, current_drop_path_keep_prob))
                    # Drop path random mask for memory space kmeans cross-attention.
                    memory_kmeans_attention_drop_path_mask = (
                        drop_path.generate_drop_path_random_mask(
                            memory_space_output, current_drop_path_keep_prob))
                    # Drop path random mask for memory space feed-forward network.
                    memory_space_feed_forward_network_drop_path_mask = (
                        drop_path.generate_drop_path_random_mask(
                            memory_space_output, current_drop_path_keep_prob))
                    transformer_block_fn_input_list += [
                        pixel_space_drop_path_mask,
                        memory_space_attention_drop_path_mask,
                        memory_kmeans_attention_drop_path_mask,
                        memory_space_feed_forward_network_drop_path_mask]

                # Build the sub-layers when the transformer_block_fn is called for the
                # first time. Otherwise, recompute_grad will not track newly built model
                # parameters.
                if self._first_building_call:
                    _ = transformer_block_fn_no_recompute(
                        tuple(transformer_block_fn_input_list))
                # Apply a dual-path transformer.
                features, memory_space_output, auxiliary_outputs = (
                    transformer_block_fn(tuple(transformer_block_fn_input_list)))

        # Now the first call has finished and the sub-layers have been built.
        self._first_building_call = False
        # We also return the non-activated output so that the function is compatible
        # with a decoder that takes a non-activated tensor as input.
        if return_auxiliary_outputs:
            return features, memory_space_output, auxiliary_outputs
        else:
            return features, memory_space_output



def fix_gelu_issue(method):
    """
    Torch 1.12 added an approximate parameter that our old models dont have.
    Monkey patch it in.

    References:
        https://github.com/pytorch/pytorch/pull/61439
    """
    for name, mod in method.named_modules():
        if mod.__class__.__name__ == "GELU":
            if not hasattr(mod, "approximate"):
                mod.approximate = "none"


def fix_package_modules():
    # Monkey Patch torch.package
    import sys

    if sys.version_info[0:2] >= (3, 10):
        try:
            from torch.package import _stdlib

            _stdlib._get_stdlib_modules = lambda: sys.stdlib_module_names
        except Exception:
            pass


def add_safe_globals():
    """
    Stop gap to allow for loading of common classes now that torch load
    disallows de-serialization of unknown classes. Safely handles cases where
    attributes might not exist in different Python/NumPy versions.

    References:
        https://github.com/huggingface/transformers/pull/34632
    """
    import torch
    import kwcoco
    import numpy as np

    # Safely get multiarray module
    try:
        np_core = np._core  # Newer NumPy versions
    except AttributeError:
        np_core = np.core  # Older NumPy versions

    # Build safe globals list dynamically
    safe_globals = []

    # Add kwcoco.CategoryTree if available
    safe_globals.append(kwcoco.category_tree.CategoryTree)

    # Add numpy multiarray reconstruct if available
    safe_globals.append(np_core.multiarray._reconstruct)
    safe_globals.append(np_core.multiarray.scalar)

    # Add basic numpy types
    safe_globals.extend([
        np.ndarray,
        np.dtype,
    ])

    # Add numpy dtype classes if they exist
    dtype_classes = [
        'UInt8DType', 'UInt16DType', 'UInt32DType', 'UInt64DType',
        'Int8DType', 'Int16DType', 'Int32DType', 'Int64DType',
        'Float16DType', 'Float32DType', 'Float64DType',
        'Complex64DType', 'Complex128DType',
        'BoolDType'
    ]

    for dtype_name in dtype_classes:
        try:
            dtype_class = getattr(np.dtypes, dtype_name)
            safe_globals.append(dtype_class)
        except AttributeError:
            continue

    # Only try to add safe globals if we found any and torch supports it
    if safe_globals:
        try:
            torch.serialization.add_safe_globals(safe_globals)
        except AttributeError:
            pass



"""
CommandLine:
    HAS_DVC=1 xdoctest geowatch/tasks/depth_pcd/model_test.py __doc__


Example:
    >>> # xdoctest: +REQUIRES(env:HAS_DVC)
    >>> import numpy as np
    >>> import geowatch
    >>> import ubelt as ub
    >>> from geowatch.tasks.depth_pcd.model import getModel
    >>> model = getModel()
    >>> expt_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
    >>> model.load_weights(expt_dvc_dpath + '/models/depth_pcd/basicModel2.h5')
    >>> out = model.predict(np.zeros((1,400,400,3)))
    >>> shapes = [o.shape for o in out]
    >>> print('shapes = {}'.format(ub.urepr(shapes, nl=1)))
"""


def mwe_tensorflow():
    r"""
    Small example that tests if tensorflow will raise a DNN error in this env
    or not.

    References:
        https://www.tensorflow.org/install/pip

    Check CuDNN version

        !apt-cache policy libcudnn8


    Debugging:
        # Try running this example in the minimum pyenv311 env before
        # installing geowatch
        docker run \
            --gpus all  \
            --volume "$HOME"/.cache/pip:/root/.cache/pip \
            -it pyenv:311 \
            bash

        # pip install tensorflow ipython nvidia-cudnn-cu11
        pip install tensorflow=="2.12.0" nvidia-cudnn-cu11==8.6.0.163

        python -c "if 1:
            import tensorflow as tf
            print(tf.config.list_physical_devices())
            from tensorflow.keras.models import Model
            conv = tf.keras.models.Sequential([
                tf.keras.layers.Conv2D(64, (3, 3), activation='relu', input_shape=(28, 28, 1)),
            ])
            i = tf.keras.Input([28, 28, 1], batch_size=1)
            out = conv(i)
            model = Model(inputs=i, outputs=[out])
            import numpy as np
            model.predict(np.zeros((1, 28, 28, 1)))
        "

    """

    import tensorflow as tf
    from tensorflow.keras.models import Model
    conv = tf.keras.models.Sequential([
        tf.keras.layers.Conv2D(64, (3, 3), activation='relu', input_shape=(28, 28, 1)),
    ])

    i = tf.keras.Input([28, 28, 1], batch_size=1)
    out = conv(i)

    model = Model(inputs=i, outputs=[out])
    import numpy as np
    out = model.predict(np.zeros((1, 28, 28, 1)))



from torch import optim
from lop.algos.convGnT import ConvGnT
import torch.nn.functional as F
from lop.utils.AdamGnT import AdamGnT


class ConvCBP(object):
    """
    The Continual Backprop algorithm
    """
    def __init__(self, net, step_size=0.001, loss='mse', opt='sgd', beta=0.9, beta_2=0.999, replacement_rate=0.0001,
                 decay_rate=0.9, init='kaiming', util_type='contribution', maturity_threshold=100, device='cpu',
                 momentum=0, weight_decay=0):
        self.net = net

        # define the optimizer
        if opt == 'sgd':
            self.opt = optim.SGD(self.net.parameters(), lr=step_size, momentum=momentum, weight_decay=weight_decay)
        elif opt == 'adam':
            self.opt = AdamGnT(self.net.parameters(), lr=step_size, betas=(beta, beta_2), weight_decay=weight_decay)

        # define the loss function
        self.loss_func = {'nll': F.cross_entropy, 'mse': F.mse_loss}[loss]

        # a placeholder
        self.previous_features = None

        # define the generate-and-test object for the given network
        self.gnt = ConvGnT(
            net=self.net.layers,
            hidden_activation=self.net.act_type,
            opt=self.opt,
            replacement_rate=replacement_rate,
            decay_rate=decay_rate,
            init=init,
            num_last_filter_outputs=net.last_filter_output,
            util_type=util_type,
            maturity_threshold=maturity_threshold,
            device=device,
        )

    def learn(self, x, target):
        """
        Learn using one step of gradient-descent and generate-&-test
        :param x: input
        :param target: desired output
        :return: loss
        """
        # do a forward pass and get the hidden activations
        output, features = self.net.predict(x=x)
        loss = self.loss_func(output, target)
        self.previous_features = features

        # do the backward pass and take a gradient step
        loss.backward()
        self.opt.step()
        self.opt.zero_grad()

        # take a generate-and-test step
        self.gnt.gen_and_test(features=self.previous_features)

        return loss.detach(), output
